# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import contextlib

import funsor
from funsor.adjoint import AdjointTape
from funsor.sum_product import _partition

from pyro.contrib.funsor import to_data, to_funsor
from pyro.contrib.funsor.handlers import enum, plate, replay, trace
from pyro.contrib.funsor.infer.elbo import ELBO, Jit_ELBO
from pyro.distributions.util import copy_docs_from
from pyro.infer import TraceEnum_ELBO as _OrigTraceEnum_ELBO


# Work around a bug in unfold_contraction_generic_tuple interacting with
# Approximate introduced in https://github.com/pyro-ppl/funsor/pull/488 .
# Once fixed, this can be replaced by funsor.optimizer.apply_optimizer().
def apply_optimizer(x):
    with funsor.interpretations.normalize:
        expr = funsor.interpreter.reinterpret(x)

    with funsor.optimizer.optimize_base:
        return funsor.interpreter.reinterpret(expr)


def terms_from_trace(tr):
    """Helper function to extract elbo components from execution traces."""
    # data structure containing densities, measures, scales, and identification
    # of free variables as either product (plate) variables or sum (measure) variables
    terms = {
        "log_factors": [],
        "log_measures": [],
        "scale": to_funsor(1.0),
        "plate_vars": frozenset(),
        "measure_vars": frozenset(),
        "plate_to_step": dict(),
    }
    for name, node in tr.nodes.items():
        # add markov dimensions to the plate_to_step dictionary
        if node["type"] == "markov_chain":
            terms["plate_to_step"][node["name"]] = node["value"]
            # ensure previous step variables are added to measure_vars
            for step in node["value"]:
                terms["measure_vars"] |= frozenset(
                    {
                        var
                        for var in step[1:-1]
                        if tr.nodes[var]["funsor"].get("log_measure", None) is not None
                    }
                )
        if (
            node["type"] != "sample"
            or type(node["fn"]).__name__ == "_Subsample"
            or node["infer"].get("_do_not_score", False)
        ):
            continue
        # grab plate dimensions from the cond_indep_stack
        terms["plate_vars"] |= frozenset(
            f.name for f in node["cond_indep_stack"] if f.vectorized
        )
        # grab the log-measure, found only at sites that are not replayed or observed
        if node["funsor"].get("log_measure", None) is not None:
            terms["log_measures"].append(node["funsor"]["log_measure"])
            # sum (measure) variables: the fresh non-plate variables at a site
            terms["measure_vars"] |= (
                frozenset(node["funsor"]["value"].inputs) | {name}
            ) - terms["plate_vars"]
        # grab the scale, assuming a common subsampling scale
        if (
            node.get("replay_active", False)
            and set(node["funsor"]["log_prob"].inputs) & terms["measure_vars"]
            and float(to_data(node["funsor"]["scale"])) != 1.0
        ):
            # model site that depends on enumerated variable: common scale
            terms["scale"] = node["funsor"]["scale"]
        else:  # otherwise: default scale behavior
            node["funsor"]["log_prob"] = (
                node["funsor"]["log_prob"] * node["funsor"]["scale"]
            )
        # grab the log-density, found at all sites except those that are not replayed
        if node["is_observed"] or not node.get("replay_skipped", False):
            terms["log_factors"].append(node["funsor"]["log_prob"])
    # add plate dimensions to the plate_to_step dictionary
    terms["plate_to_step"].update(
        {plate: terms["plate_to_step"].get(plate, {}) for plate in terms["plate_vars"]}
    )
    return terms


@copy_docs_from(_OrigTraceEnum_ELBO)
class TraceMarkovEnum_ELBO(ELBO):
    def differentiable_loss(self, model, guide, *args, **kwargs):
        # get batched, enumerated, to_funsor-ed traces from the guide and model
        with (
            plate(size=self.num_particles)
            if self.num_particles > 1
            else contextlib.ExitStack()
        ), enum(
            first_available_dim=(
                (-self.max_plate_nesting - 1) if self.max_plate_nesting else None
            )
        ):
            guide_tr = trace(guide).get_trace(*args, **kwargs)
            model_tr = trace(replay(model, trace=guide_tr)).get_trace(*args, **kwargs)

        # extract from traces all metadata that we will need to compute the elbo
        guide_terms = terms_from_trace(guide_tr)
        model_terms = terms_from_trace(model_tr)

        # guide side enumeration is not supported
        if any(guide_terms["plate_to_step"].values()):
            raise NotImplementedError(
                "TraceMarkovEnum_ELBO does not yet support guide side Markov enumeration"
            )

        # build up a lazy expression for the elbo
        with funsor.terms.lazy:
            # identify and contract out auxiliary variables in the model with partial_sum_product
            contracted_factors, uncontracted_factors = [], []
            for f in model_terms["log_factors"]:
                if model_terms["measure_vars"].intersection(f.inputs):
                    contracted_factors.append(f)
                else:
                    uncontracted_factors.append(f)
            # incorporate the effects of subsampling and handlers.scale through a common scale factor
            markov_dims = frozenset(
                {plate for plate, step in model_terms["plate_to_step"].items() if step}
            )
            contracted_costs = [
                model_terms["scale"] * f
                for f in funsor.sum_product.dynamic_partial_sum_product(
                    funsor.ops.logaddexp,
                    funsor.ops.add,
                    model_terms["log_measures"] + contracted_factors,
                    plate_to_step=model_terms["plate_to_step"],
                    eliminate=model_terms["measure_vars"] | markov_dims,
                )
            ]

            costs = contracted_costs + uncontracted_factors  # model costs: logp
            costs += [-f for f in guide_terms["log_factors"]]  # guide costs: -logq

            # finally, integrate out guide variables in the elbo and all plates
            plate_vars = guide_terms["plate_vars"] | model_terms["plate_vars"]
            elbo = to_funsor(0, output=funsor.Real)
            for cost in costs:
                # compute the marginal logq in the guide corresponding to this cost term
                log_prob = funsor.sum_product.sum_product(
                    funsor.ops.logaddexp,
                    funsor.ops.add,
                    guide_terms["log_measures"],
                    plates=plate_vars,
                    eliminate=(plate_vars | guide_terms["measure_vars"])
                    - frozenset(cost.inputs),
                )
                # compute the expected cost term E_q[logp] or E_q[-logq] using the marginal logq for q
                elbo_term = funsor.Integrate(
                    log_prob, cost, guide_terms["measure_vars"] & frozenset(cost.inputs)
                )
                elbo += elbo_term.reduce(
                    funsor.ops.add, plate_vars & frozenset(cost.inputs)
                )

        # evaluate the elbo, using memoize to share tensor computation where possible
        with funsor.interpretations.memoize():
            return -to_data(apply_optimizer(elbo))


@copy_docs_from(_OrigTraceEnum_ELBO)
class TraceEnum_ELBO(ELBO):
    def differentiable_loss(self, model, guide, *args, **kwargs):
        # get batched, enumerated, to_funsor-ed traces from the guide and model
        with (
            plate(size=self.num_particles)
            if self.num_particles > 1
            else contextlib.ExitStack()
        ), enum(
            first_available_dim=(
                (-self.max_plate_nesting - 1) if self.max_plate_nesting else None
            )
        ):
            guide_tr = trace(guide).get_trace(*args, **kwargs)
            model_tr = trace(replay(model, trace=guide_tr)).get_trace(*args, **kwargs)

        # extract from traces all metadata that we will need to compute the elbo
        guide_terms = terms_from_trace(guide_tr)
        model_terms = terms_from_trace(model_tr)

        # build up a lazy expression for the elbo
        with funsor.terms.lazy:
            # identify and contract out auxiliary variables in the model with partial_sum_product
            contracted_factors, uncontracted_factors = [], []
            for f in model_terms["log_factors"]:
                if model_terms["measure_vars"].intersection(f.inputs):
                    contracted_factors.append(f)
                else:
                    uncontracted_factors.append(f)
            contracted_costs = []
            # incorporate the effects of subsampling and handlers.scale through a common scale factor
            for group_factors, group_vars in _partition(
                model_terms["log_measures"] + contracted_factors,
                model_terms["measure_vars"],
            ):
                group_factor_vars = frozenset().union(
                    *[f.inputs for f in group_factors]
                )
                group_plates = model_terms["plate_vars"] & group_factor_vars
                outermost_plates = frozenset.intersection(
                    *(frozenset(f.inputs) & group_plates for f in group_factors)
                )
                elim_plates = group_plates - outermost_plates
                for f in funsor.sum_product.partial_sum_product(
                    funsor.ops.logaddexp,
                    funsor.ops.add,
                    group_factors,
                    plates=group_plates,
                    eliminate=group_vars | elim_plates,
                ):
                    contracted_costs.append(model_terms["scale"] * f)

            # accumulate costs from model (logp) and guide (-logq)
            costs = contracted_costs + uncontracted_factors  # model costs: logp
            costs += [-f for f in guide_terms["log_factors"]]  # guide costs: -logq

            # compute expected cost
            # Cf. pyro.infer.util.Dice.compute_expectation()
            # https://github.com/pyro-ppl/pyro/blob/0.3.0/pyro/infer/util.py#L212
            # TODO Replace this with funsor.Expectation
            plate_vars = guide_terms["plate_vars"] | model_terms["plate_vars"]
            # compute the marginal logq in the guide corresponding to each cost term
            targets = dict()
            for cost in costs:
                input_vars = frozenset(cost.inputs)
                if input_vars not in targets:
                    targets[input_vars] = funsor.Tensor(
                        funsor.ops.new_zeros(
                            funsor.tensor.get_default_prototype(),
                            tuple(v.size for v in cost.inputs.values()),
                        ),
                        cost.inputs,
                        cost.dtype,
                    )
            with AdjointTape() as tape:
                logzq = funsor.sum_product.sum_product(
                    funsor.ops.logaddexp,
                    funsor.ops.add,
                    guide_terms["log_measures"] + list(targets.values()),
                    plates=plate_vars,
                    eliminate=(plate_vars | guide_terms["measure_vars"]),
                )
            marginals = tape.adjoint(
                funsor.ops.logaddexp, funsor.ops.add, logzq, tuple(targets.values())
            )
            # finally, integrate out guide variables in the elbo and all plates
            elbo = to_funsor(0, output=funsor.Real)
            for cost in costs:
                target = targets[frozenset(cost.inputs)]
                logzq_local = marginals[target].reduce(
                    funsor.ops.logaddexp, frozenset(cost.inputs) - plate_vars
                )
                log_prob = marginals[target] - logzq_local
                elbo_term = funsor.Integrate(
                    log_prob,
                    cost,
                    guide_terms["measure_vars"] & frozenset(log_prob.inputs),
                )
                elbo += elbo_term.reduce(
                    funsor.ops.add, plate_vars & frozenset(cost.inputs)
                )

        # evaluate the elbo, using memoize to share tensor computation where possible
        with funsor.interpretations.memoize():
            return -to_data(apply_optimizer(elbo))


class JitTraceEnum_ELBO(Jit_ELBO, TraceEnum_ELBO):
    pass


class JitTraceMarkovEnum_ELBO(Jit_ELBO, TraceMarkovEnum_ELBO):
    pass



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

"""Tasks such as choosing the next question to ask in a psychology study, designing an election polling
strategy, and deciding which compounds to synthesize and test in biological sciences are all fundamentally asking the
same question: how do we design an experiment to maximize the information gathered?  Pyro is designed to support
automated optimal experiment design: specifying a model and guide is enough to obtain optimal designs for many different
kinds of experiment scenarios. Check out our experimental design tutorials that use Pyro to
`design an adaptive psychology study <https://pyro.ai/examples/working_memory.html>`_ that uses past data to
select the next question, and `design an election polling strategy <https://pyro.ai/examples/elections.html>`_ that
aims to give the strongest prediction about the eventual winner of the election.

Bayesian optimal experimental design (BOED) is a powerful methodology for tackling experimental design problems and
is the framework adopted by Pyro.
In the BOED framework, we begin with a Bayesian model with a likelihood :math:`p(y|\\theta,d)` and a prior
:math:`p(\\theta)` on the target latent variables. In Pyro, any fully Bayesian model can be used in the BOED framework.
The sample sites corresponding to experimental outcomes are the *observation* sites, those corresponding to
latent variables of interest are the *target* sites. The design :math:`d` is the argument to the model, and is not
a random variable.

In the BOED framework, we choose the design that optimizes the expected information gain (EIG) on the targets
:math:`\\theta` from running the experiment

    :math:`\\text{EIG}(d) = \\mathbf{E}_{p(y|d)} [H[p(\\theta)] − H[p(\\theta|y, d)]]` ,

where :math:`H[·]` represents the entropy and :math:`p(\\theta|y, d) \\propto p(\\theta)p(y|\\theta, d)` is the
posterior we get from
running the experiment with design :math:`d` and observing :math:`y`. In other words, the optimal design is the one
that, in expectation over possible future observations, most reduces posterior entropy
over the target latent variables. If the predictive model is correct, this forms a design strategy that is
(one-step) optimal from an information-theoretic viewpoint. For further details, see [1, 2].

The :mod:`pyro.contrib.oed` module provides tools to create optimal experimental
designs for Pyro models. In particular, it provides estimators for the
expected information gain (EIG).

To estimate the EIG for a particular design, we first set up our Pyro model. For example::

    def model(design):

        # This line allows batching of designs, treating all batch dimensions as independent
        with pyro.plate_stack("plate_stack", design.shape):

            # We use a Normal prior for theta
            theta = pyro.sample("theta", dist.Normal(torch.tensor(0.0), torch.tensor(1.0)))

            # We use a simple logistic regression model for the likelihood
            logit_p = theta - design
            y = pyro.sample("y", dist.Bernoulli(logits=logit_p))

            return y

We then select an appropriate EIG estimator, such as::

    eig = nmc_eig(model, design, observation_labels=["y"], target_labels=["theta"], N=2500, M=50)

It is possible to estimate the EIG across a grid of designs::

    designs = torch.stack([design1, design2], dim=0)

to find the best design from a number of options.

[1] Chaloner, Kathryn, and Isabella Verdinelli. "Bayesian experimental design: A review."
Statistical Science (1995): 273-304.

[2] Foster, Adam, et al. "Variational Bayesian Optimal Experimental Design." arXiv preprint arXiv:1903.05480 (2019).

"""

from pyro.contrib.oed import eig, search

__all__ = ["search", "eig"]



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math
from collections import OrderedDict

import torch

import pyro
import pyro.distributions as dist
from pyro.infer.mcmc.mcmc_kernel import MCMCKernel
from pyro.infer.mcmc.util import initialize_model


class RandomWalkKernel(MCMCKernel):
    r"""
    Simple gradient-free kernel that utilizes an isotropic gaussian random walk in the unconstrained
    latent space of the model. The step size that controls the variance of the kernel is adapted during
    warm-up with a simple adaptation scheme that targets a user-provided acceptance probability.

    :param model: Python callable containing Pyro primitives.
    :param float init_step_size: A positive float that controls the initial step size. Defaults to 0.1.
    :param float target_accept_prob: The target acceptance probability used during adaptation of
        the step size. Defaults to 0.234.

    Example:

        >>> true_coefs = torch.tensor([1., 2., 3.])
        >>> data = torch.randn(2000, 3)
        >>> dim = 3
        >>> labels = dist.Bernoulli(logits=(true_coefs * data).sum(-1)).sample()
        >>>
        >>> def model(data):
        ...     coefs_mean = torch.zeros(dim)
        ...     coefs = pyro.sample('beta', dist.Normal(coefs_mean, torch.ones(3)))
        ...     y = pyro.sample('y', dist.Bernoulli(logits=(coefs * data).sum(-1)), obs=labels)
        ...     return y
        >>>
        >>> hmc_kernel = RandomWalkKernel(model, init_step_size=0.2)
        >>> mcmc = MCMC(hmc_kernel, num_samples=200, warmup_steps=100)
        >>> mcmc.run(data)
        >>> mcmc.get_samples()['beta'].mean(0)  # doctest: +SKIP
        tensor([ 0.9819,  1.9258,  2.9737])
    """

    def __init__(
        self, model, init_step_size: float = 0.1, target_accept_prob: float = 0.234
    ):
        if not isinstance(init_step_size, float) or init_step_size <= 0.0:
            raise ValueError("init_step_size must be a positive float.")

        if (
            not isinstance(target_accept_prob, float)
            or target_accept_prob <= 0.0
            or target_accept_prob >= 1.0
        ):
            raise ValueError(
                "target_accept_prob must be a float in the interval (0, 1)."
            )

        self.model = model
        self.init_step_size = init_step_size
        self.target_accept_prob = target_accept_prob

        self._t = 0
        self._log_step_size = math.log(init_step_size)
        self._accept_cnt = 0
        self._mean_accept_prob = 0.0
        super().__init__()

    def setup(self, warmup_steps, *args, **kwargs):
        self._warmup_steps = warmup_steps
        (
            self._initial_params,
            self.potential_fn,
            self.transforms,
            self._prototype_trace,
        ) = initialize_model(
            self.model,
            model_args=args,
            model_kwargs=kwargs,
        )
        self._energy_last = self.potential_fn(self._initial_params)

    def sample(self, params):
        step_size = math.exp(self._log_step_size)
        new_params = {
            k: v + step_size * torch.randn(v.shape, dtype=v.dtype, device=v.device)
            for k, v in params.items()
        }
        energy_proposal = self.potential_fn(new_params)
        delta_energy = energy_proposal - self._energy_last

        accept_prob = (-delta_energy).exp().clamp(max=1.0).item()
        rand = pyro.sample(
            "rand_t={}".format(self._t),
            dist.Uniform(0.0, 1.0),
        )
        accepted = False
        if rand < accept_prob:
            accepted = True
            params = new_params
            self._energy_last = energy_proposal

        if self._t <= self._warmup_steps:
            adaptation_speed = max(0.001, 0.1 / math.sqrt(1 + self._t))
            self._log_step_size += adaptation_speed * (
                accept_prob - self.target_accept_prob
            )

        self._t += 1

        if self._t > self._warmup_steps:
            n = self._t - self._warmup_steps
            if accepted:
                self._accept_cnt += 1
        else:
            n = self._t

        self._mean_accept_prob += (accept_prob - self._mean_accept_prob) / n

        return params.copy()

    @property
    def initial_params(self):
        return self._initial_params

    @initial_params.setter
    def initial_params(self, params):
        self._initial_params = params

    def logging(self):
        return OrderedDict(
            [
                ("step size", "{:.2e}".format(math.exp(self._log_step_size))),
                ("acc. prob", "{:.3f}".format(self._mean_accept_prob)),
            ]
        )

    def diagnostics(self):
        return {
            "acceptance rate": self._accept_cnt / (self._t - self._warmup_steps),
        }



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import Categorical, OneHotCategorical

from pyro.distributions import constraints
from pyro.distributions.hmm import _sequential_logmatmulexp
from pyro.distributions.torch_distribution import TorchDistribution
from pyro.distributions.util import broadcast_shape


class MissingDataDiscreteHMM(TorchDistribution):
    """
    HMM with discrete latent states and discrete observations, allowing for
    missing data or variable length sequences. Observations are assumed
    to be one hot encoded; rows with all zeros indicate missing data.

    .. warning:: Unlike in pyro's pyro.distributions.DiscreteHMM, which
        computes the probability of the first state as
        initial.T @ transition @ emission
        this distribution uses the standard HMM convention,
        initial.T @ emission

    :param ~torch.Tensor initial_logits: A logits tensor for an initial
        categorical distribution over latent states. Should have rightmost
        size ``state_dim`` and be broadcastable to
        ``(batch_size, state_dim)``.
    :param ~torch.Tensor transition_logits: A logits tensor for transition
        conditional distributions between latent states. Should have rightmost
        shape ``(state_dim, state_dim)`` (old, new), and be broadcastable
        to ``(batch_size, state_dim, state_dim)``.
    :param ~torch.Tensor observation_logits: A logits tensor for observation
        distributions from latent states. Should have rightmost shape
        ``(state_dim, categorical_size)``, where ``categorical_size`` is the
        dimension of the categorical output, and be broadcastable
        to ``(batch_size, state_dim, categorical_size)``.
    """

    arg_constraints = {
        "initial_logits": constraints.real_vector,
        "transition_logits": constraints.independent(constraints.real, 2),
        "observation_logits": constraints.independent(constraints.real, 2),
    }
    support = constraints.independent(constraints.nonnegative_integer, 2)

    def __init__(
        self, initial_logits, transition_logits, observation_logits, validate_args=None
    ):
        if initial_logits.dim() < 1:
            raise ValueError(
                "expected initial_logits to have at least one dim, "
                "actual shape = {}".format(initial_logits.shape)
            )
        if transition_logits.dim() < 2:
            raise ValueError(
                "expected transition_logits to have at least two dims, "
                "actual shape = {}".format(transition_logits.shape)
            )
        if observation_logits.dim() < 2:
            raise ValueError(
                "expected observation_logits to have at least two dims, "
                "actual shape = {}".format(transition_logits.shape)
            )
        shape = broadcast_shape(
            initial_logits.shape[:-1],
            transition_logits.shape[:-2],
            observation_logits.shape[:-2],
        )
        if len(shape) == 0:
            shape = torch.Size([1])
        batch_shape = shape
        event_shape = (1, observation_logits.shape[-1])
        self.initial_logits = initial_logits - initial_logits.logsumexp(-1, True)
        self.transition_logits = transition_logits - transition_logits.logsumexp(
            -1, True
        )
        self.observation_logits = observation_logits - observation_logits.logsumexp(
            -1, True
        )
        super(MissingDataDiscreteHMM, self).__init__(
            batch_shape, event_shape, validate_args=validate_args
        )

    def log_prob(self, value):
        """
        :param ~torch.Tensor value: One-hot encoded observation. Must be
            real-valued (float) and broadcastable to
            ``(batch_size, num_steps, categorical_size)`` where
            ``categorical_size`` is the dimension of the categorical output.
            Missing data is represented by zeros, i.e.
            ``value[batch, step, :] == tensor([0, ..., 0])``.
            Variable length observation sequences can be handled by padding
            the sequence with zeros at the end.
        """

        assert value.shape[-1] == self.event_shape[1]

        # Combine observation and transition factors.
        value_logits = torch.matmul(
            value, torch.transpose(self.observation_logits, -2, -1)
        )
        result = self.transition_logits.unsqueeze(-3) + value_logits[..., 1:, None, :]

        # Eliminate time dimension.
        result = _sequential_logmatmulexp(result)

        # Combine initial factor.
        result = self.initial_logits + value_logits[..., 0, :] + result.logsumexp(-1)

        # Marginalize out final state.
        result = result.logsumexp(-1)
        return result

    def sample(self, sample_shape=torch.Size([])):
        """
        :param ~torch.Size sample_shape: Sample shape, last dimension must be
            ``num_steps`` and must be broadcastable to
            ``(batch_size, num_steps)``. batch_size must be int not tuple.
        """
        # shape: batch_size x num_steps x categorical_size
        shape = broadcast_shape(
            torch.Size(list(self.batch_shape) + [1, 1]),
            torch.Size(list(sample_shape) + [1]),
            torch.Size((1, 1, self.event_shape[-1])),
        )
        # state: batch_size x state_dim
        state = OneHotCategorical(logits=self.initial_logits).sample()
        # sample: batch_size x num_steps x categorical_size
        sample = torch.zeros(shape)
        for i in range(shape[-2]):
            # batch_size x 1 x state_dim @
            # batch_size x state_dim x categorical_size
            obs_logits = torch.matmul(
                state.unsqueeze(-2), self.observation_logits
            ).squeeze(-2)
            sample[:, i, :] = OneHotCategorical(logits=obs_logits).sample()
            # batch_size x 1 x state_dim @
            # batch_size x state_dim x state_dim
            trans_logits = torch.matmul(
                state.unsqueeze(-2), self.transition_logits
            ).squeeze(-2)
            state = OneHotCategorical(logits=trans_logits).sample()

        return sample

    def filter(self, value):
        """
        Compute the marginal probability of the state variable at each
        step conditional on the previous observations.

        :param ~torch.Tensor value: One-hot encoded observation.
            Must be real-valued (float) and broadcastable to
            ``(batch_size, num_steps, categorical_size)`` where
            ``categorical_size`` is the dimension of the categorical output.
        """
        # batch_size x num_steps x state_dim
        shape = broadcast_shape(
            torch.Size(list(self.batch_shape) + [1, 1]),
            torch.Size(list(value.shape[:-1]) + [1]),
            torch.Size((1, 1, self.initial_logits.shape[-1])),
        )
        filter = torch.zeros(shape)

        # Combine observation and transition factors.
        # batch_size x num_steps x state_dim
        value_logits = torch.matmul(
            value, torch.transpose(self.observation_logits, -2, -1)
        )
        # batch_size x num_steps-1 x state_dim x state_dim
        result = self.transition_logits.unsqueeze(-3) + value_logits[..., 1:, None, :]

        # Forward pass. (This could be parallelized using the
        # Sarkka & Garcia-Fernandez method.)
        filter[..., 0, :] = self.initial_logits + value_logits[..., 0, :]
        filter[..., 0, :] = filter[..., 0, :] - torch.logsumexp(
            filter[..., 0, :], -1, True
        )
        for i in range(1, shape[-2]):
            filter[..., i, :] = torch.logsumexp(
                filter[..., i - 1, :, None] + result[..., i - 1, :, :], -2
            )
            filter[..., i, :] = filter[..., i, :] - torch.logsumexp(
                filter[..., i, :], -1, True
            )
        return filter

    def smooth(self, value):
        """
        Compute posterior expected value of state at each position (smoothing).

        :param ~torch.Tensor value: One-hot encoded observation.
            Must be real-valued (float) and broadcastable to
            ``(batch_size, num_steps, categorical_size)`` where
            ``categorical_size`` is the dimension of the categorical output.
        """
        # Compute filter and initialize.
        filter = self.filter(value)
        shape = filter.shape
        backfilter = torch.zeros(shape)

        # Combine observation and transition factors.
        # batch_size x num_steps x state_dim
        value_logits = torch.matmul(
            value, torch.transpose(self.observation_logits, -2, -1)
        )
        # batch_size x num_steps-1 x state_dim x state_dim
        result = self.transition_logits.unsqueeze(-3) + value_logits[..., 1:, None, :]
        # Construct backwards filter.
        for i in range(shape[-2] - 1, 0, -1):
            backfilter[..., i - 1, :] = torch.logsumexp(
                backfilter[..., i, None, :] + result[..., i - 1, :, :], -1
            )

        # Compute smoothed version.
        smooth = filter + backfilter
        smooth = smooth - torch.logsumexp(smooth, -1, True)
        return smooth

    def sample_states(self, value):
        """
        Sample states with forward filtering-backward sampling algorithm.

        :param ~torch.Tensor value: One-hot encoded observation.
            Must be real-valued (float) and broadcastable to
            ``(batch_size, num_steps, categorical_size)`` where
            ``categorical_size`` is the dimension of the categorical output.
        """
        filter = self.filter(value)
        shape = filter.shape
        joint = filter.unsqueeze(-1) + self.transition_logits.unsqueeze(-3)
        states = torch.zeros(shape[:-1], dtype=torch.long)
        states[..., -1] = Categorical(logits=filter[..., -1, :]).sample()
        for i in range(shape[-2] - 1, 0, -1):
            logits = torch.gather(
                joint[..., i - 1, :, :],
                -1,
                states[..., i, None, None]
                * torch.ones([shape[-1], 1], dtype=torch.long),
            ).squeeze(-1)
            states[..., i - 1] = Categorical(logits=logits).sample()
        return states

    def map_states(self, value):
        """
        Compute maximum a posteriori (MAP) estimate of state variable with
        Viterbi algorithm.

        :param ~torch.Tensor value: One-hot encoded observation.
            Must be real-valued (float) and broadcastable to
            ``(batch_size, num_steps, categorical_size)`` where
            ``categorical_size`` is the dimension of the categorical output.
        """
        # Setup for Viterbi.
        # batch_size x num_steps x state_dim
        shape = broadcast_shape(
            torch.Size(list(self.batch_shape) + [1, 1]),
            torch.Size(list(value.shape[:-1]) + [1]),
            torch.Size((1, 1, self.initial_logits.shape[-1])),
        )
        state_logits = torch.zeros(shape)
        state_traceback = torch.zeros(shape, dtype=torch.long)

        # Combine observation and transition factors.
        # batch_size x num_steps x state_dim
        value_logits = torch.matmul(
            value, torch.transpose(self.observation_logits, -2, -1)
        )
        # batch_size x num_steps-1 x state_dim x state_dim
        result = self.transition_logits.unsqueeze(-3) + value_logits[..., 1:, None, :]

        # Forward pass.
        state_logits[..., 0, :] = self.initial_logits + value_logits[..., 0, :]
        for i in range(1, shape[-2]):
            transit_weights = (
                state_logits[..., i - 1, :, None] + result[..., i - 1, :, :]
            )
            state_logits[..., i, :], state_traceback[..., i, :] = torch.max(
                transit_weights, -2
            )
        # Traceback.
        map_states = torch.zeros(shape[:-1], dtype=torch.long)
        map_states[..., -1] = torch.argmax(state_logits[..., -1, :], -1)
        for i in range(shape[-2] - 1, 0, -1):
            map_states[..., i - 1] = torch.gather(
                state_traceback[..., i, :], -1, map_states[..., i].unsqueeze(-1)
            ).squeeze(-1)
        return map_states

    def given_states(self, states):
        """
        Distribution conditional on the state variable.

        :param ~torch.Tensor map_states: State trajectory. Must be
            integer-valued (long) and broadcastable to
            ``(batch_size, num_steps)``.
        """
        shape = broadcast_shape(
            list(self.batch_shape) + [1, 1],
            list(states.shape[:-1]) + [1, 1],
            [1, 1, self.observation_logits.shape[-1]],
        )
        states_index = states.unsqueeze(-1) * torch.ones(shape, dtype=torch.long)
        obs_logits = self.observation_logits * torch.ones(shape)
        logits = torch.gather(obs_logits, -2, states_index)
        return OneHotCategorical(logits=logits)

    def sample_given_states(self, states):
        """
        Sample an observation conditional on the state variable.

        :param ~torch.Tensor map_states: State trajectory. Must be
            integer-valued (long) and broadcastable to
            ``(batch_size, num_steps)``.
        """
        conditional = self.given_states(states)
        return conditional.sample()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math
from abc import ABCMeta, abstractmethod

import torch
from torch.distributions import biject_to

import pyro
from pyro import poutine
from pyro.distributions import Delta
from pyro.distributions.util import copy_docs_from
from pyro.infer.autoguide.guides import AutoContinuous
from pyro.infer.autoguide.initialization import init_to_sample
from pyro.infer.trace_elbo import Trace_ELBO


def vectorize(fn, num_particles, max_plate_nesting):
    def _fn(*args, **kwargs):
        with pyro.plate(
            "num_particles_vectorized", num_particles, dim=-max_plate_nesting - 1
        ):
            return fn(*args, **kwargs)

    return _fn


class _SVGDGuide(AutoContinuous):
    """
    This modification of :class:`AutoContinuous` is used internally in the
    :class:`SVGD` inference algorithm.
    """

    def __init__(self, model):
        super().__init__(model, init_loc_fn=init_to_sample)

    def get_posterior(self, *args, **kwargs):
        svgd_particles = pyro.param("svgd_particles", self._init_loc)
        return Delta(svgd_particles, event_dim=1)


class SteinKernel(object, metaclass=ABCMeta):
    """
    Abstract class for kernels used in the :class:`SVGD` inference algorithm.
    """

    @abstractmethod
    def log_kernel_and_grad(self, particles):
        """
        Compute the component kernels and their gradients.

        :param particles: a tensor with shape (N, D)
        :returns: A pair (`log_kernel`, `kernel_grad`) where `log_kernel` is a (N, N, D)-shaped
            tensor equal to the logarithm of the kernel and `kernel_grad` is a (N, N, D)-shaped
            tensor where the entry (n, m, d) represents the derivative of `log_kernel` w.r.t.
            x_{m,d}, where x_{m,d} is the d^th dimension of particle m.
        """
        raise NotImplementedError


@copy_docs_from(SteinKernel)
class RBFSteinKernel(SteinKernel):
    """
    A RBF kernel for use in the SVGD inference algorithm. The bandwidth of the kernel is chosen from the
    particles using a simple heuristic as in reference [1].

    :param float bandwidth_factor: Optional factor by which to scale the bandwidth, defaults to 1.0.
    :ivar float ~.bandwidth_factor: Property that controls the factor by which to scale the bandwidth
        at each iteration.

    References

    [1] "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm,"
        Qiang Liu, Dilin Wang
    """

    def __init__(self, bandwidth_factor=None):
        """
        :param float bandwidth_factor: Optional factor by which to scale the bandwidth
        """
        self.bandwidth_factor = bandwidth_factor

    def _bandwidth(self, norm_sq):
        """
        Compute the bandwidth along each dimension using the median pairwise squared distance between particles.
        """
        num_particles = norm_sq.size(0)
        index = torch.arange(num_particles)
        norm_sq = norm_sq[index > index.unsqueeze(-1), ...]
        median = norm_sq.median(dim=0)[0]
        if self.bandwidth_factor is not None:
            median = self.bandwidth_factor * median
        assert median.shape == norm_sq.shape[-1:]
        return median / math.log(num_particles + 1)

    @torch.no_grad()
    def log_kernel_and_grad(self, particles):
        delta_x = particles.unsqueeze(0) - particles.unsqueeze(1)  # N N D
        assert delta_x.dim() == 3
        norm_sq = delta_x.pow(2.0)  # N N D
        h = self._bandwidth(norm_sq)  # D
        log_kernel = -(norm_sq / h)  # N N D
        grad_term = 2.0 * delta_x / h  # N N D
        assert log_kernel.shape == grad_term.shape
        return log_kernel, grad_term

    @property
    def bandwidth_factor(self):
        return self._bandwidth_factor

    @bandwidth_factor.setter
    def bandwidth_factor(self, bandwidth_factor):
        """
        :param float bandwidth_factor: Optional factor by which to scale the bandwidth
        """
        if bandwidth_factor is not None:
            assert bandwidth_factor > 0.0, "bandwidth_factor must be positive."
        self._bandwidth_factor = bandwidth_factor


@copy_docs_from(SteinKernel)
class IMQSteinKernel(SteinKernel):
    r"""
    An IMQ (inverse multi-quadratic) kernel for use in the SVGD inference algorithm [1]. The bandwidth of the kernel
    is chosen from the particles using a simple heuristic as in reference [2]. The kernel takes the form

    :math:`K(x, y) = (\alpha + ||x-y||^2/h)^{\beta}`

    where :math:`\alpha` and :math:`\beta` are user-specified parameters and :math:`h` is the bandwidth.

    :param float alpha: Kernel hyperparameter, defaults to 0.5.
    :param float beta: Kernel hyperparameter, defaults to -0.5.
    :param float bandwidth_factor: Optional factor by which to scale the bandwidth, defaults to 1.0.
    :ivar float ~.bandwidth_factor: Property that controls the factor by which to scale the bandwidth
        at each iteration.

    References

    [1] "Stein Points," Wilson Ye Chen, Lester Mackey, Jackson Gorham, Francois-Xavier Briol, Chris. J. Oates.
    [2] "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm," Qiang Liu, Dilin Wang
    """

    def __init__(self, alpha=0.5, beta=-0.5, bandwidth_factor=None):
        """
        :param float alpha: Kernel hyperparameter, defaults to 0.5.
        :param float beta: Kernel hyperparameter, defaults to -0.5.
        :param float bandwidth_factor: Optional factor by which to scale the bandwidth
        """
        assert alpha > 0.0, "alpha must be positive."
        assert beta < 0.0, "beta must be negative."
        self.alpha = alpha
        self.beta = beta
        self.bandwidth_factor = bandwidth_factor

    def _bandwidth(self, norm_sq):
        """
        Compute the bandwidth along each dimension using the median pairwise squared distance between particles.
        """
        num_particles = norm_sq.size(0)
        index = torch.arange(num_particles)
        norm_sq = norm_sq[index > index.unsqueeze(-1), ...]
        median = norm_sq.median(dim=0)[0]
        if self.bandwidth_factor is not None:
            median = self.bandwidth_factor * median
        assert median.shape == norm_sq.shape[-1:]
        return median / math.log(num_particles + 1)

    @torch.no_grad()
    def log_kernel_and_grad(self, particles):
        delta_x = particles.unsqueeze(0) - particles.unsqueeze(1)  # N N D
        assert delta_x.dim() == 3
        norm_sq = delta_x.pow(2.0)  # N N D
        h = self._bandwidth(norm_sq)  # D
        base_term = self.alpha + norm_sq / h
        log_kernel = self.beta * torch.log(base_term)  # N N D
        grad_term = (-2.0 * self.beta) * delta_x / h  # N N D
        grad_term = grad_term / base_term
        assert log_kernel.shape == grad_term.shape
        return log_kernel, grad_term

    @property
    def bandwidth_factor(self):
        return self._bandwidth_factor

    @bandwidth_factor.setter
    def bandwidth_factor(self, bandwidth_factor):
        """
        :param float bandwidth_factor: Optional factor by which to scale the bandwidth
        """
        if bandwidth_factor is not None:
            assert bandwidth_factor > 0.0, "bandwidth_factor must be positive."
        self._bandwidth_factor = bandwidth_factor


class SVGD:
    """
    A basic implementation of Stein Variational Gradient Descent as described in reference [1].

    :param model: The model (callable containing Pyro primitives). Model must be fully vectorized
        and may only contain continuous latent variables.
    :param kernel: a SVGD compatible kernel like :class:`RBFSteinKernel`.
    :param optim: A wrapper for a PyTorch optimizer.
    :type optim: pyro.optim.PyroOptim
    :param int num_particles: The number of particles used in SVGD.
    :param int max_plate_nesting: The max number of nested :func:`pyro.plate` contexts in the model.
    :param str mode: Whether to use a Kernelized Stein Discrepancy that makes use of `multivariate`
        test functions (as in [1]) or `univariate` test functions (as in [2]). Defaults to `univariate`.

    Example usage:

    .. code-block:: python

        from pyro.infer import SVGD, RBFSteinKernel
        from pyro.optim import Adam

        kernel = RBFSteinKernel()
        adam = Adam({"lr": 0.1})
        svgd = SVGD(model, kernel, adam, num_particles=50, max_plate_nesting=0)

        for step in range(500):
            svgd.step(model_arg1, model_arg2)

        final_particles = svgd.get_named_particles()

    References

    [1] "Stein Variational Gradient Descent: A General Purpose Bayesian Inference Algorithm,"
        Qiang Liu, Dilin Wang
    [2] "Kernelized Complete Conditional Stein Discrepancy,"
        Raghav Singhal, Saad Lahlou, Rajesh Ranganath
    """

    def __init__(
        self, model, kernel, optim, num_particles, max_plate_nesting, mode="univariate"
    ):
        assert callable(model)
        assert isinstance(kernel, SteinKernel), "Must provide a valid SteinKernel"
        assert isinstance(
            optim, pyro.optim.PyroOptim
        ), "Must provide a valid Pyro optimizer"
        assert num_particles > 1, "Must use at least two particles"
        assert max_plate_nesting >= 0
        assert mode in [
            "univariate",
            "multivariate",
        ], "mode must be one of (univariate, multivariate)"

        self.model = vectorize(model, num_particles, max_plate_nesting)
        self.kernel = kernel
        self.optim = optim
        self.num_particles = num_particles
        self.max_plate_nesting = max_plate_nesting
        self.mode = mode

        self.loss = Trace_ELBO().differentiable_loss
        self.guide = _SVGDGuide(self.model)

    def get_named_particles(self):
        """
        Create a dictionary mapping name to vectorized value, of the form ``{name: tensor}``.
        The leading dimension of each tensor corresponds to particles, i.e. this creates a struct of arrays.
        """
        return {
            site["name"]: biject_to(site["fn"].support)(unconstrained_value)
            for site, unconstrained_value in self.guide._unpack_latent(
                pyro.param("svgd_particles")
            )
        }

    @torch.no_grad()
    def step(self, *args, **kwargs):
        """
        Computes the SVGD gradient, passing args and kwargs to the model,
        and takes a gradient step.

        :return dict: A dictionary of the form {name: float}, where each float
            is a mean squared gradient. This can be used to monitor the convergence of SVGD.
        """
        # compute gradients of log model joint
        with torch.enable_grad(), poutine.trace(param_only=True) as param_capture:
            loss = self.loss(self.model, self.guide, *args, **kwargs)
            loss.backward()

        # get particles used in the _SVGDGuide and reshape to have num_particles leading dimension
        particles = pyro.param("svgd_particles").unconstrained()
        reshaped_particles = particles.reshape(self.num_particles, -1)
        reshaped_particles_grad = particles.grad.reshape(self.num_particles, -1)

        # compute kernel ingredients
        log_kernel, kernel_grad = self.kernel.log_kernel_and_grad(reshaped_particles)

        if self.mode == "multivariate":
            kernel = log_kernel.sum(-1).exp()
            assert kernel.shape == (self.num_particles, self.num_particles)
            attractive_grad = torch.mm(kernel, reshaped_particles_grad)
            repulsive_grad = torch.einsum("nm,nm...->n...", kernel, kernel_grad)
        elif self.mode == "univariate":
            kernel = log_kernel.exp()
            assert kernel.shape == (
                self.num_particles,
                self.num_particles,
                reshaped_particles.size(-1),
            )
            attractive_grad = torch.einsum(
                "nmd,md->nd", kernel, reshaped_particles_grad
            )
            repulsive_grad = torch.einsum("nmd,nmd->nd", kernel, kernel_grad)

        # combine the attractive and repulsive terms in the SVGD gradient
        assert attractive_grad.shape == repulsive_grad.shape
        particles.grad = (attractive_grad + repulsive_grad).reshape(
            particles.shape
        ) / self.num_particles

        # compute per-parameter mean squared gradients
        squared_gradients = {
            site["name"]: value.mean().item()
            for site, value in self.guide._unpack_latent(particles.grad.pow(2.0))
        }

        # torch.optim objects gets instantiated for any params that haven't been seen yet
        params = set(
            site["value"].unconstrained() for site in param_capture.trace.nodes.values()
        )
        self.optim(params)

        # zero gradients
        pyro.infer.util.zero_grads(params)

        # return per-parameter mean squared gradients to user
        return squared_gradients



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from pyro.infer import SVI, config_enumerate  # noqa: F401

from .discrete import infer_discrete  # noqa: F401
from .elbo import ELBO  # noqa: F401
from .trace_elbo import JitTrace_ELBO, Trace_ELBO  # noqa: F401
from .traceenum_elbo import (  # noqa: F401
    JitTraceEnum_ELBO,
    JitTraceMarkovEnum_ELBO,
    TraceEnum_ELBO,
    TraceMarkovEnum_ELBO,
)
from .tracetmc_elbo import JitTraceTMC_ELBO, TraceTMC_ELBO  # noqa: F401



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

"""
This file contains helpers to use `Zuko <https://zuko.readthedocs.io/>`_-based
normalizing flows within Pyro piplines.

Accompanying tutorials can be found at `tutorial/svi_flow_guide.ipynb` and
`tutorial/vae_flow_prior.ipynb`.
"""

import torch
from torch import Size, Tensor

import pyro


class ZukoToPyro(pyro.distributions.TorchDistribution):
    r"""Wraps a Zuko distribution as a Pyro distribution.

    If ``dist`` has an ``rsample_and_log_prob`` method, like Zuko's flows, it will be
    used when sampling instead of ``rsample``. The returned log density will be cached
    for later scoring.

    :param dist: A distribution instance.
    :type dist: torch.distributions.Distribution

    .. code-block:: python

        flow = zuko.flows.MAF(features=5)

        # flow() is a torch.distributions.Distribution

        dist = flow()
        x = dist.sample((2, 3))
        log_p = dist.log_prob(x)

        # ZukoToPyro(flow()) is a pyro.distributions.Distribution

        dist = ZukoToPyro(flow())
        x = dist((2, 3))
        log_p = dist.log_prob(x)

        with pyro.plate("data", 42):
            z = pyro.sample("z", dist)
    """

    def __init__(self, dist: torch.distributions.Distribution):
        self.dist = dist
        self.cache = {}

    @property
    def has_rsample(self) -> bool:
        return self.dist.has_rsample

    @property
    def event_shape(self) -> Size:
        return self.dist.event_shape

    @property
    def batch_shape(self) -> Size:
        return self.dist.batch_shape

    def __call__(self, shape: Size = ()) -> Tensor:
        if hasattr(self.dist, "rsample_and_log_prob"):  # fast sampling + scoring
            x, self.cache[x] = self.dist.rsample_and_log_prob(shape)
        elif self.has_rsample:
            x = self.dist.rsample(shape)
        else:
            x = self.dist.sample(shape)

        return x

    def log_prob(self, x: Tensor) -> Tensor:
        if x in self.cache:
            return self.cache[x]
        else:
            return self.dist.log_prob(x)

    def expand(self, *args, **kwargs):
        return ZukoToPyro(self.dist.expand(*args, **kwargs))



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

import torch

SymmArrowhead = namedtuple("SymmArrowhead", ["top", "bottom_diag"])
TriuArrowhead = namedtuple("TriuArrowhead", ["top", "bottom_diag"])


def sqrt(x):
    """
    EXPERIMENTAL Computes the upper triangular square root of an
    symmetric arrowhead matrix.

    :param SymmArrowhead x: an symmetric arrowhead matrix
    :return: the square root of `x`
    :rtype: TriuArrowhead
    """
    assert isinstance(x, SymmArrowhead)
    head_size = x.top.size(0)
    if head_size == 0:
        return TriuArrowhead(x.top, x.bottom_diag.sqrt())

    A, B = x.top[:, :head_size], x.top[:, head_size:]
    # NB: the complexity is O(N * head_size^2)
    # ref: https://en.wikipedia.org/wiki/Schur_complement#Background
    Dsqrt = x.bottom_diag.sqrt()

    # On cholesky error, retry with smaller tail part B.
    num_attempts = 6
    for i in range(num_attempts):
        B_Dsqrt = B / Dsqrt.unsqueeze(-2)  # shape: head_size x N
        schur_complement = A - B_Dsqrt.matmul(
            B_Dsqrt.t()
        )  # complexity: head_size^2 x N
        # we will decompose schur_complement to U @ U.T (so that the sqrt matrix
        # is upper triangular) using some `flip` operators:
        #   flip(cholesky(flip(schur_complement)))
        try:
            top_left = torch.flip(
                torch.linalg.cholesky(torch.flip(schur_complement, (-2, -1))), (-2, -1)
            )
            break
        except RuntimeError:
            B = B / 2
            continue
        raise RuntimeError(
            "Singular schur complement in computing Cholesky of the input"
            " arrowhead matrix"
        )

    top_right = B_Dsqrt
    top = torch.cat([top_left, top_right], -1)
    bottom_diag = Dsqrt
    return TriuArrowhead(top, bottom_diag)


def triu_inverse(x):
    """
    EXPERIMENTAL Computes the inverse of an upper-triangular arrowhead matrix.

    :param TriuArrowhead x: an upper-triangular arrowhead matrix.
    :return: the inverse of `x`
    :rtype: TriuArrowhead
    """
    assert isinstance(x, TriuArrowhead)
    head_size = x.top.size(0)
    if head_size == 0:
        return TriuArrowhead(x.top, x.bottom_diag.reciprocal())

    A, B = x.top[:, :head_size], x.top[:, head_size:]
    B_Dinv = B / x.bottom_diag.unsqueeze(-2)

    identity = torch.eye(head_size, dtype=A.dtype, device=A.device)
    top_left = torch.linalg.solve_triangular(A, identity, upper=True)
    top_right = -top_left.matmul(B_Dinv)  # complexity: head_size^2 x N
    top = torch.cat([top_left, top_right], -1)
    bottom_diag = x.bottom_diag.reciprocal()
    return TriuArrowhead(top, bottom_diag)


def triu_matvecmul(x, y, transpose=False):
    """
    EXPERIMENTAL Computes matrix-vector product of an upper-triangular
    arrowhead matrix `x` and a vector `y`.

    :param TriuArrowhead x: an upper-triangular arrowhead matrix.
    :param torch.Tensor y: a 1D tensor
    :return: matrix-vector product of `x` and `y`
    :rtype: TriuArrowhead
    """
    assert isinstance(x, TriuArrowhead)
    head_size = x.top.size(0)
    if transpose:
        z = x.top.transpose(-2, -1).matmul(y[:head_size])
        # here we exploit the diagonal structure of the bottom right part
        # of arrowhead_sqrt matrix; so the complexity is still O(N)
        top = z[:head_size]
        bottom = z[head_size:] + x.bottom_diag * y[head_size:]
    else:
        top = x.top.matmul(y)
        bottom = x.bottom_diag * y[head_size:]
    return torch.cat([top, bottom], 0)


def triu_gram(x):
    """
    EXPERIMENTAL Computes the gram matrix `x.T @ x` from an upper-triangular
    arrowhead matrix `x`.

    :param TriuArrowhead x: an upper-triangular arrowhead matrix.
    :return: the square of `x`
    :rtype: TriuArrowhead
    """
    assert isinstance(x, TriuArrowhead)
    head_size = x.top.size(0)
    if head_size == 0:
        return x.bottom_diag.pow(2)

    A, B = x.top[:, :head_size], x.top[:, head_size:]
    top = A.t().matmul(x.top)
    bottom_left = top[:, head_size:].t()
    # the following matmul operator is O(N^2 x head_size)
    bottom_right = B.t().matmul(B) + x.bottom_diag.pow(2).diag()
    return torch.cat([top, torch.cat([bottom_left, bottom_right], -1)], 0)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions.utils import broadcast_all

from pyro.distributions import Normal, constraints
from pyro.distributions.torch_distribution import TorchDistribution

LOG_ROOT_TWO_PI = 0.5 * math.log(2.0 * math.pi)


class GroupedNormalNormal(TorchDistribution):
    r"""
    This likelihood, which operates on groups of real-valued scalar observations, is obtained by
    integrating out a latent mean for each group. Both the prior on each latent mean as well as the
    observation likelihood for each data point are univariate Normal distributions.
    The prior means are controlled by `prior_loc` and `prior_scale`. The observation noise of the
    Normal likelihood is controlled by `obs_scale`, which is allowed to vary from observation to
    observation. The tensor of indices `group_idx` connects each observation to one of the groups
    specified by `prior_loc` and `prior_scale`.

    See e.g. Eqn. (55) in ref. [1] for relevant expressions in a simpler case with scalar `obs_scale`.

    Example:

    >>> num_groups = 3
    >>> num_data = 4
    >>> prior_loc = torch.randn(num_groups)
    >>> prior_scale = torch.rand(num_groups)
    >>> obs_scale = torch.rand(num_data)
    >>> group_idx = torch.tensor([1, 0, 2, 1]).long()
    >>> values = torch.randn(num_data)
    >>> gnn = GroupedNormalNormal(prior_loc, prior_scale, obs_scale, group_idx)
    >>> assert gnn.log_prob(values).shape == ()

    References:
    [1] "Conjugate Bayesian analysis of the Gaussian distribution," Kevin P. Murphy.

    :param torch.Tensor prior_loc: Tensor of shape `(num_groups,)` specifying the prior mean of the latent
        of each group.
    :param torch.Tensor prior_scale: Tensor of shape `(num_groups,)` specifying the prior scale of the latent
        of each group.
    :param torch.Tensor obs_scale: Tensor of shape `(num_data,)` specifying the scale of the observation noise
        of each observation.
    :param torch.LongTensor group_idx: Tensor of indices of shape `(num_data,)` linking each observation to one
        of the `num_groups` groups that are specified in `prior_loc` and `prior_scale`.
    """

    arg_constraints = {
        "prior_loc": constraints.real,
        "prior_scale": constraints.positive,
        "obs_scale": constraints.positive,
    }
    support = constraints.real

    def __init__(
        self, prior_loc, prior_scale, obs_scale, group_idx, validate_args=None
    ):
        if prior_loc.ndim not in [0, 1] or prior_scale.ndim not in [0, 1]:
            raise ValueError(
                "prior_loc and prior_scale must be broadcastable to 1D tensors of the same shape."
            )

        if obs_scale.ndim not in [0, 1]:
            raise ValueError(
                "obs_scale must be broadcastable to a 1-dimensional tensor."
            )

        if group_idx.ndim != 1 or not isinstance(group_idx, torch.LongTensor):
            raise ValueError("group_idx must be a 1-dimensional tensor of indices.")

        prior_loc, prior_scale = broadcast_all(prior_loc, prior_scale)
        obs_scale, group_idx = broadcast_all(obs_scale, group_idx)

        self.prior_loc = prior_loc
        self.prior_scale = prior_scale
        self.obs_scale = obs_scale
        self.group_idx = group_idx
        batch_shape = prior_loc.shape[:-1]

        if batch_shape != torch.Size([]):
            raise ValueError("GroupedNormalNormal only supports trivial batch_shape's.")

        self.num_groups = prior_loc.size(0)
        if group_idx.min().item() < 0 or group_idx.max().item() >= self.num_groups:
            raise ValueError(
                "Each index in group_idx must be an integer in the inclusive range [0, prior_loc.size(0) - 1]."
            )

        self.num_data_per_batch = prior_loc.new_zeros(self.num_groups).scatter_add(
            0, self.group_idx, prior_loc.new_ones(self.group_idx.shape)
        )
        super().__init__(batch_shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        raise NotImplementedError

    def sample(self, sample_shape=()):
        raise NotImplementedError

    def get_posterior(self, value):
        """
        Get a `pyro.distributions.Normal` distribution that encodes the posterior distribution
        over the vector of latents specified by `prior_loc` and `prior_scale` conditioned on the
        observed data specified by `value`.
        """
        if value.shape != self.group_idx.shape:
            raise ValueError(
                "GroupedNormalNormal.get_posterior only supports values that have the same shape as group_idx."
            )

        obs_scale_sq_inv = self.obs_scale.pow(-2)
        prior_scale_sq_inv = self.prior_scale.pow(-2)

        obs_scale_sq_inv_sum = torch.zeros_like(self.prior_loc).scatter_add(
            0, self.group_idx, obs_scale_sq_inv
        )
        precision = prior_scale_sq_inv + obs_scale_sq_inv_sum
        scaled_value_sum = torch.zeros_like(self.prior_loc).scatter_add(
            0, self.group_idx, value * obs_scale_sq_inv
        )

        loc = (scaled_value_sum + self.prior_loc * prior_scale_sq_inv) / precision
        scale = precision.rsqrt()

        return Normal(loc=loc, scale=scale)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)

        group_idx = self.group_idx

        if value.shape != group_idx.shape:
            raise ValueError(
                "GroupedNormalNormal.log_prob only supports values that have the same shape as group_idx."
            )

        prior_scale_sq = self.prior_scale.pow(2.0)
        obs_scale_sq_inv = self.obs_scale.pow(-2)
        obs_scale_sq_inv_sum = torch.zeros_like(self.prior_loc).scatter_add(
            0, self.group_idx, obs_scale_sq_inv
        )

        scale_ratio = prior_scale_sq * obs_scale_sq_inv_sum
        delta = value - self.prior_loc[group_idx]
        scaled_delta = delta * obs_scale_sq_inv
        scaled_delta_sum = torch.zeros_like(self.prior_loc).scatter_add(
            0, self.group_idx, scaled_delta
        )

        result1 = -(self.num_data_per_batch * LOG_ROOT_TWO_PI).sum()
        result2 = -0.5 * torch.log1p(scale_ratio).sum() - self.obs_scale.log().sum()
        result3 = -0.5 * torch.dot(delta, scaled_delta)
        numerator = prior_scale_sq * scaled_delta_sum.pow(2)
        result4 = 0.5 * (numerator / (1.0 + scale_ratio)).sum()

        return result1 + result2 + result3 + result4



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from typing import Union

from torch import Tensor

from pyro.distributions import TransformedDistribution
from pyro.distributions.transforms import (
    AbsTransform,
    AffineTransform,
    ExpTransform,
    PowerTransform,
    SigmoidTransform,
    SoftmaxTransform,
    TanhTransform,
    Transform,
)


class RVMagicOps:
    """Mixin class for overloading __magic__ operations on random variables."""

    def __add__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(x, 1))
        )

    def __radd__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(x, 1))
        )

    def __sub__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(-x, 1))
        )

    def __rsub__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(x, -1))
        )

    def __mul__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(0, x))
        )

    def __rmul__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(0, x))
        )

    def __truediv__(self, x: Union[float, Tensor]):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(0, 1 / x))
        )

    def __neg__(self):
        return RandomVariable(
            TransformedDistribution(self.distribution, AffineTransform(0, -1))
        )

    def __abs__(self):
        return RandomVariable(
            TransformedDistribution(self.distribution, AbsTransform())
        )

    def __pow__(self, x):
        return RandomVariable(
            TransformedDistribution(self.distribution, PowerTransform(x))
        )


class RVChainOps:
    """Mixin class for performing common unary/binary operations on/between
    random variables/constant tensors using method chaining syntax.
    """

    def add(self, x):
        return self + x

    def sub(self, x):
        return self - x

    def mul(self, x):
        return self * x

    def div(self, x):
        return self / x

    def abs(self):
        return abs(self)

    def pow(self, x):
        return self**x

    def neg(self):
        return -self

    def exp(self):
        return self.transform(ExpTransform())

    def log(self):
        return self.transform(ExpTransform().inv)

    def sigmoid(self):
        return self.transform(SigmoidTransform())

    def tanh(self):
        return self.transform(TanhTransform())

    def softmax(self):
        return self.transform(SoftmaxTransform())


class RandomVariable(RVMagicOps, RVChainOps):
    """EXPERIMENTAL random variable container class around a distribution

    Representation of a distribution interpreted as a random variable. Rather
    than directly manipulating a probability density by applying pointwise
    transformations to it, this allows for simple arithmetic transformations of
    the random variable the distribution represents. For more flexibility,
    consider using the `transform` method. Note that if you perform a
    non-invertible transform (like `abs(X)` or `X**2`), certain things might
    not work properly.

    Can switch between `RandomVariable` and `Distribution` objects with the
    convenient `Distribution.rv` and `RandomVariable.dist` properties.

    Supports either chaining operations or arithmetic operator overloading.

    Example usage::

        # This should be equivalent to an Exponential distribution.
        RandomVariable(Uniform(0, 1)).log().neg().dist

        # These two distributions Y1, Y2 should be the same
        X = Uniform(0, 1).rv
        Y1 = X.mul(4).pow(0.5).sub(1).abs().neg().dist
        Y2 = (-abs((4*X)**(0.5) - 1)).dist
    """

    def __init__(self, distribution):
        """Wraps a distribution as a RandomVariable

        :param distribution: The `Distribution` object to wrap
        :type distribution: ~pyro.distributions.distribution.Distribution
        """
        self.distribution = distribution

    def transform(self, t: Transform):
        """Performs a transformation on the distribution underlying the RV.

        :param t: The transformation (or sequence of transformations) to be
            applied to the distribution. There are many examples to be found in
            `torch.distributions.transforms` and `pyro.distributions.transforms`,
            or you can subclass directly from `Transform`.
        :type t: ~pyro.distributions.transforms.Transform

        :return: The transformed `RandomVariable`
        :rtype: ~pyro.contrib.randomvariable.random_variable.RandomVariable
        """
        dist = TransformedDistribution(self.distribution, t)
        return RandomVariable(dist)

    @property
    def dist(self):
        """Convenience property for exposing the distribution underlying the
        random variable.

        :return: The `Distribution` object underlying the random variable
        :rtype: ~pyro.distributions.distribution.Distribution
        """
        return self.distribution



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import contextlib

import funsor

from pyro.contrib.funsor import to_data
from pyro.contrib.funsor.handlers import enum, plate, replay, trace
from pyro.contrib.funsor.infer.elbo import ELBO, Jit_ELBO
from pyro.contrib.funsor.infer.traceenum_elbo import apply_optimizer, terms_from_trace
from pyro.distributions.util import copy_docs_from
from pyro.infer import TraceTMC_ELBO as _OrigTraceTMC_ELBO


@copy_docs_from(_OrigTraceTMC_ELBO)
class TraceTMC_ELBO(ELBO):
    def differentiable_loss(self, model, guide, *args, **kwargs):
        with (
            plate(size=self.num_particles)
            if self.num_particles > 1
            else contextlib.ExitStack()
        ), enum(
            first_available_dim=(
                (-self.max_plate_nesting - 1) if self.max_plate_nesting else None
            )
        ):
            guide_tr = trace(guide).get_trace(*args, **kwargs)
            model_tr = trace(replay(model, trace=guide_tr)).get_trace(*args, **kwargs)

        model_terms = terms_from_trace(model_tr)
        guide_terms = terms_from_trace(guide_tr)

        log_measures = guide_terms["log_measures"] + model_terms["log_measures"]
        log_factors = model_terms["log_factors"] + [
            -f for f in guide_terms["log_factors"]
        ]
        plate_vars = model_terms["plate_vars"] | guide_terms["plate_vars"]
        measure_vars = model_terms["measure_vars"] | guide_terms["measure_vars"]

        with funsor.terms.lazy:
            elbo = funsor.sum_product.sum_product(
                funsor.ops.logaddexp,
                funsor.ops.add,
                log_measures + log_factors,
                eliminate=measure_vars | plate_vars,
                plates=plate_vars,
            )

        return -to_data(apply_optimizer(elbo))


class JitTraceTMC_ELBO(Jit_ELBO, TraceTMC_ELBO):
    pass



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import constraints
from torch.nn import Parameter

from pyro.contrib.gp.kernels.kernel import Kernel
from pyro.nn.module import PyroParam


class Coregionalize(Kernel):
    r"""
    A kernel for the linear model of coregionalization
    :math:`k(x,z) = x^T (W W^T + D) z` where :math:`W` is an
    ``input_dim``-by-``rank`` matrix and typically ``rank < input_dim``,
    and ``D`` is a diagonal matrix.

    This generalizes the
    :class:`~pyro.contrib.gp.kernels.dot_product.Linear` kernel to multiple
    features with a low-rank-plus-diagonal weight matrix. The typical use case
    is for modeling correlations among outputs of a multi-output GP, where
    outputs are coded as distinct data points with one-hot coded features
    denoting which output each datapoint represents.

    If only ``rank`` is specified, the kernel ``(W W^T + D)`` will be
    randomly initialized to a matrix with expected value the identity matrix.

    References:

    [1] Mauricio A. Alvarez, Lorenzo Rosasco, Neil D. Lawrence (2012)
        Kernels for Vector-Valued Functions: a Review

    :param int input_dim: Number of feature dimensions of inputs.
    :param int rank: Optional rank. This is only used if ``components`` is
        unspecified. If neigher ``rank`` nor ``components`` is specified,
        then ``rank`` defaults to ``input_dim``.
    :param torch.Tensor components: An optional ``(input_dim, rank)`` shaped
        matrix that maps features to ``rank``-many components. If unspecified,
        this will be randomly initialized.
    :param torch.Tensor diagonal: An optional vector of length ``input_dim``.
        If unspecified, this will be set to constant ``0.5``.
    :param list active_dims: List of feature dimensions of the input which the
        kernel acts on.
    :param str name: Name of the kernel.
    """

    def __init__(
        self, input_dim, rank=None, components=None, diagonal=None, active_dims=None
    ):
        super().__init__(input_dim, active_dims)

        # Add a low-rank kernel with expected value torch.eye(input_dim, input_dim) / 2.
        if components is None:
            rank = input_dim if rank is None else rank
            components = torch.randn(input_dim, rank) * (0.5 / rank) ** 0.5
        else:
            rank = components.size(-1)
        if components.shape != (input_dim, rank):
            raise ValueError(
                "Expected components.shape == ({},rank), actual {}".format(
                    input_dim, components.shape
                )
            )
        self.components = Parameter(components)

        # Add a diagonal component initialized to torch.eye(input_dim, input_dim) / 2,
        # such that the total kernel has expected value the identity matrix.
        diagonal = (
            components.new_ones(input_dim) * 0.5 if diagonal is None else diagonal
        )
        if diagonal.shape != (input_dim,):
            raise ValueError(
                "Expected diagonal.shape == ({},), actual {}".format(
                    input_dim, diagonal.shape
                )
            )
        self.diagonal = PyroParam(diagonal, constraints.positive)

    def forward(self, X, Z=None, diag=False):
        X = self._slice_input(X)
        Xc = X.matmul(self.components)

        if diag:
            return (Xc**2).sum(-1) + (X**2).mv(self.diagonal)

        if Z is None:
            Z = X
            Zc = Xc
        else:
            Z = self._slice_input(Z)
            Zc = Z.matmul(self.components)

        return Xc.matmul(Zc.t()) + (X * self.diagonal).matmul(Z.t())



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

from pyroapi import *  # noqa F401
from pyroapi import __all__  # noqa F401

warnings.warn("pyro.generic has moved to the pyroapi package", DeprecationWarning)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import numbers
from functools import singledispatch

import torch
from torch.distributions import transform_to, transforms

import pyro.distributions as dist
from pyro.infer.reparam import DiscreteCosineReparam, HaarReparam
from pyro.poutine.messenger import Messenger
from pyro.poutine.util import site_is_subsample
from pyro.primitives import get_param_store


def time_reparam_dct(msg):
    """
    EXPERIMENTAL Configures ``poutine.reparam()`` to use a ``HaarReparam`` for
    all sites inside the ``time`` plate.
    """
    if msg["is_observed"]:
        return
    for frame in msg["cond_indep_stack"]:
        if frame.name == "time":
            dim = frame.dim - msg["fn"].event_dim
            return HaarReparam(dim=dim, experimental_allow_batch=True)


def time_reparam_haar(msg):
    """
    EXPERIMENTAL Configures ``poutine.reparam()`` to use a ``DiscreteCosineReparam`` for
    all sites inside the ``time`` plate.
    """
    if msg["is_observed"]:
        return
    for frame in msg["cond_indep_stack"]:
        if frame.name == "time":
            dim = frame.dim - msg["fn"].event_dim
            return DiscreteCosineReparam(dim=dim, experimental_allow_batch=True)


class MarkDCTParamMessenger(Messenger):
    """
    EXPERIMENTAL Messenger to mark DCT dimension of parameter, for use with
    :class:`pyro.optim.optim.DCTAdam`.

    :param str name: The name of the plate along which to apply discrete cosine
        transforms on gradients.
    """

    def __init__(self, name):
        super().__init__()
        self.name = name

    def _postprocess_message(self, msg):
        if msg["type"] != "param":
            return
        event_dim = msg["kwargs"].get("event_dim")
        if event_dim is None:
            return
        for frame in msg["cond_indep_stack"]:
            if frame.name == self.name:
                value = msg["value"]
                event_dim += value.unconstrained().dim() - value.dim()
                value.unconstrained()._pyro_dct_dim = frame.dim - event_dim
                return


class PrefixWarmStartMessenger(Messenger):
    """
    EXPERIMENTAL Assuming the global param store has been populated with params
    defined on a short time window, re-initialize by splicing old params with
    new initial params defined on a longer time window.
    """

    def _pyro_param(self, msg):
        store = get_param_store()
        name = msg["name"]
        if name not in store:
            return

        if len(msg["args"]) >= 2:
            new = msg["args"][1]
        elif "init_tensor" in msg["kwargs"]:
            new = msg["kwargs"]["init_tensor"]
        else:
            return  # no init tensor specified

        if callable(new):
            new = new()
        old = store[name]
        assert new.dim() == old.dim()
        if new.shape == old.shape:
            return

        # Splice old (warm start) and new (init) tensors.
        # This only works for time-homogeneous constraints.
        t = transform_to(store._constraints[name])
        new = t.inv(new)
        old = t.inv(old)
        for dim in range(new.dim()):
            if new.size(dim) != old.size(dim):
                break
        assert new.size(dim) > old.size(dim)
        assert new.shape[dim + 1 :] == old.shape[dim + 1 :]
        split = old.size(dim)
        index = (slice(None),) * dim + (slice(split, None),)
        new = torch.cat([old, new[index]], dim=dim)
        store[name] = t(new)


class PrefixReplayMessenger(Messenger):
    """
    EXPERIMENTAL Given a trace of training data, replay a model with batched
    sites extended to include both training and forecast time, using the guide
    trace for the training prefix and samples from the prior on the forecast
    postfix.

    :param trace: a guide trace.
    :type trace: ~pyro.poutine.trace_struct.Trace
    """

    def __init__(self, trace):
        super().__init__()
        self.trace = trace

    def _pyro_post_sample(self, msg):
        if site_is_subsample(msg):
            return

        name = msg["name"]
        if name not in self.trace:
            return

        model_value = msg["value"]
        guide_value = self.trace.nodes[name]["value"]
        if model_value.shape == guide_value.shape:
            msg["value"] = guide_value
            return

        # Search for a single dim with mismatched size.
        assert model_value.dim() == guide_value.dim()
        for dim in range(model_value.dim()):
            if model_value.size(dim) != guide_value.size(dim):
                break
        assert model_value.size(dim) > guide_value.size(dim)
        assert model_value.shape[dim + 1 :] == guide_value.shape[dim + 1 :]
        split = guide_value.size(dim)
        index = (slice(None),) * dim + (slice(split, None),)
        msg["value"] = torch.cat([guide_value, model_value[index]], dim=dim)


class PrefixConditionMessenger(Messenger):
    """
    EXPERIMENTAL Given a prefix of t-many observations, condition a (t+f)-long
    distribution on the observations, converting it to an f-long distribution.

    :param dict data: A dict mapping site name to tensors of observations.
    """

    def __init__(self, data):
        super().__init__()
        self.data = data

    def _pyro_sample(self, msg):
        if msg["name"] not in self.data:
            return

        assert msg["value"] is None
        data = self.data[msg["name"]]
        msg["fn"] = prefix_condition(msg["fn"], data)


# ----------------------------------------------------------------------------
# The pattern-matching code in the remainder of this file could be eventually
# replace by much simpler Funsor logic.

UNIVARIATE_DISTS = {
    dist.Bernoulli: ("logits",),
    dist.Beta: ("concentration1", "concentration0"),
    dist.BetaBinomial: ("concentration1", "concentration0"),
    dist.Cauchy: ("loc", "scale"),
    dist.Dirichlet: ("concentration",),
    dist.DirichletMultinomial: ("concentration",),
    dist.Exponential: ("rate",),
    dist.Gamma: ("concentration", "rate"),
    dist.GammaPoisson: ("concentration", "rate"),
    dist.Geometric: ("logits",),
    dist.InverseGamma: ("concentration", "rate"),
    dist.Laplace: ("loc", "scale"),
    dist.LogNormal: ("loc", "scale"),
    dist.NegativeBinomial: ("total_count", "logits"),
    dist.Normal: ("loc", "scale"),
    dist.Poisson: ("rate",),
    dist.Stable: ("stability", "skew", "scale", "loc"),
    dist.StudentT: ("df", "loc", "scale"),
    dist.Uniform: ("low", "high"),
    dist.ZeroInflatedPoisson: ("rate", "gate"),
    dist.ZeroInflatedNegativeBinomial: ("total_count", "logits", "gate"),
}


@singledispatch
def prefix_condition(d, data):
    """
    EXPERIMENTAL Given a distribution ``d`` of shape ``batch_shape + (t+f, d)``
    and data ``x`` of shape ``batch_shape + (t, d)``, compute a conditional
    distribution of shape ``batch_shape + (f, d)``. Typically ``t`` is the
    number of training time steps, ``f`` is the number of forecast time steps,
    and ``d`` is the data dimension.

    :param d: a distribution with ``len(d.shape()) >= 2``
    :type d: ~pyro.distributions.Distribution
    :param data: data of dimension at least 2.
    :type data: ~torch.Tensor
    """
    try:
        return d.prefix_condition(data)
    except AttributeError as e:
        raise NotImplementedError(
            "prefix_condition() does not suport {}".format(type(d))
        ) from e


@prefix_condition.register(dist.MaskedDistribution)
def _(d, data):
    base_dist = prefix_condition(d.base_dist, data)
    mask = d._mask[tuple(slice(-size, None) for size in base_dist.batch_shape)]
    return base_dist.mask(mask)


@prefix_condition.register(dist.Independent)
def _(d, data):
    base_dist = prefix_condition(d.base_dist, data)
    return base_dist.to_event(d.reinterpreted_batch_ndims)


@prefix_condition.register(dist.IndependentHMM)
def _(d, data):
    base_data = data.transpose(-1, -2).unsqueeze(-1)
    base_dist = prefix_condition(d.base_dist, base_data)
    return dist.IndependentHMM(base_dist)


@prefix_condition.register(dist.FoldedDistribution)
def _(d, data):
    base_dist = prefix_condition(d.base_dist, data)
    return dist.FoldedDistribution(base_dist)


@prefix_condition.register(dist.TransformedDistribution)
def _(d, data):
    for t in reversed(d.transforms):
        data = t.inv(data)
    base_dist = prefix_condition(d.base_dist, data)
    return dist.TransformedDistribution(base_dist, d.transforms)


def _prefix_condition_univariate(d, data):
    t = data.size(-2)
    params = {name: getattr(d, name)[..., t:, :] for name in UNIVARIATE_DISTS[type(d)]}
    return type(d)(**params)


for _type in UNIVARIATE_DISTS:
    prefix_condition.register(_type)(_prefix_condition_univariate)


@prefix_condition.register(dist.MultivariateNormal)
def _(d, data):
    t = data.size(-2)
    loc = d.loc[..., t:, :]
    scale_tril = d.scale_tril[..., t:, :, :]
    return dist.MultivariateNormal(loc, scale_tril=scale_tril)


@singledispatch
def reshape_batch(d, batch_shape):
    """
    EXPERIMENTAL Given a distribution ``d``, reshape to different batch shape
    of same number of elements.

    This is typically used to move the the rightmost batch dimension "time" to
    an event dimension, while preserving the positions of other batch
    dimensions.

    :param d: A distribution.
    :type d: ~pyro.distributions.Distribution
    :param tuple batch_shape: A new batch shape.
    :returns: A distribution with the same type but given batch shape.
    :rtype: ~pyro.distributions.Distribution
    """
    raise NotImplementedError("reshape_batch() does not suport {}".format(type(d)))


@reshape_batch.register(dist.MaskedDistribution)
def _(d, batch_shape):
    mask = d._mask
    if not isinstance(d._mask, numbers.Number):
        mask = mask.reshape(batch_shape)
    base_dist = reshape_batch(d.base_dist, batch_shape)
    return base_dist.mask(mask)


@reshape_batch.register(dist.Independent)
def _(d, batch_shape):
    base_shape = batch_shape + d.event_shape[: d.reinterpreted_batch_ndims]
    base_dist = reshape_batch(d.base_dist, base_shape)
    return base_dist.to_event(d.reinterpreted_batch_ndims)


@reshape_batch.register(dist.IndependentHMM)
def _(d, batch_shape):
    base_shape = batch_shape + d.event_shape[-1:]
    base_dist = reshape_batch(d.base_dist, base_shape)
    return dist.IndependentHMM(base_dist)


@reshape_batch.register(dist.FoldedDistribution)
def _(d, batch_shape):
    base_dist = reshape_batch(d.base_dist, batch_shape)
    return dist.FoldedDistribution(base_dist)


@reshape_batch.register(dist.TransformedDistribution)
def _(d, batch_shape):
    base_dist = reshape_batch(d.base_dist, batch_shape)
    old_shape = d.base_dist.shape()
    new_shape = base_dist.shape()
    transforms = [
        reshape_transform_batch(t, old_shape, new_shape) for t in d.transforms
    ]
    return dist.TransformedDistribution(base_dist, transforms)


def _reshape_batch_univariate(d, batch_shape):
    batch_shape = batch_shape + (-1,) * d.event_dim
    params = {
        name: getattr(d, name).reshape(batch_shape)
        for name in UNIVARIATE_DISTS[type(d)]
    }
    return type(d)(**params)


for _type in UNIVARIATE_DISTS:
    reshape_batch.register(_type)(_reshape_batch_univariate)


@reshape_batch.register(dist.MultivariateNormal)
def _(d, batch_shape):
    dim = d.event_shape[0]
    loc = d.loc.reshape(batch_shape + (dim,))
    scale_tril = d.scale_tril.reshape(batch_shape + (dim, dim))
    return dist.MultivariateNormal(loc, scale_tril=scale_tril)


@reshape_batch.register(dist.GaussianHMM)
def _(d, batch_shape):
    init = d._init
    if init.batch_shape:
        init = init.expand(d.batch_shape)
        init = init.reshape(batch_shape)

    trans = d._trans
    if len(trans.batch_shape) > 1:
        trans = trans.expand(d.batch_shape + (-1,))
        trans = trans.reshape(batch_shape + (-1,))

    obs = d._obs
    if len(obs.batch_shape) > 1:
        obs = obs.expand(d.batch_shape + (-1,))
        obs = obs.reshape(batch_shape + (-1,))

    new = d._get_checked_instance(dist.GaussianHMM)
    new.hidden_dim = d.hidden_dim
    new.obs_dim = d.obs_dim
    new._init = init
    new._trans = trans
    new._obs = obs
    super(dist.GaussianHMM, new).__init__(
        d.duration, batch_shape, d.event_shape, validate_args=d._validate_args
    )
    return new


@reshape_batch.register(dist.LinearHMM)
def _(d, batch_shape):
    init_dist = reshape_batch(d.initial_dist, batch_shape)
    trans_mat = d.transition_matrix.reshape(
        batch_shape + (-1, d.hidden_dim, d.hidden_dim)
    )
    trans_dist = reshape_batch(d.transition_dist, batch_shape + (-1,))
    obs_mat = d.observation_matrix.reshape(batch_shape + (-1, d.hidden_dim, d.obs_dim))
    obs_dist = reshape_batch(d.observation_dist, batch_shape + (-1,))

    new = d._get_checked_instance(dist.LinearHMM)
    new.hidden_dim = d.hidden_dim
    new.obs_dim = d.obs_dim
    new.initial_dist = init_dist
    new.transition_matrix = trans_mat
    new.transition_dist = trans_dist
    new.observation_matrix = obs_mat
    new.observation_dist = obs_dist
    transforms = []
    for transform in d.transforms:
        assert type(transform) in UNIVARIATE_TRANSFORMS, (
            "Currently, reshape_batch only supports AbsTransform, "
            + "ExpTransform, SigmoidTransform transform"
        )
        old_shape = d.observation_dist.shape()
        new_shape = obs_dist.shape()
        transforms.append(reshape_transform_batch(transform, old_shape, new_shape))
    new.transforms = transforms
    super(dist.LinearHMM, new).__init__(
        d.duration, batch_shape, d.event_shape, validate_args=d._validate_args
    )
    return new


UNIVARIATE_TRANSFORMS = {
    transforms.AbsTransform: (),
    transforms.AffineTransform: ("loc", "scale"),
    transforms.ExpTransform: (),
    transforms.PowerTransform: ("exponent",),
    transforms.SigmoidTransform: (),
}


@singledispatch
def reshape_transform_batch(t, old_shape, new_shape):
    """
    EXPERIMENTAL Given a transform ``t``, reshape to different batch shape
    of same number of elements.

    This is typically used to correct the transform parameters' shapes after
    reshaping the base distribution of a transformed distribution.

    :param t: A transform.
    :type t: ~torch.distributions.transforms.Transform
    :param tuple old_shape: The current shape of ``t``.
    :param tuple new_shape: A new shape.
    :returns: A transform with the same type but given new batch shape.
    :rtype: ~torch.distributions.transforms.Transform
    """
    raise NotImplementedError(
        "reshape_transform_batch() does not suport {}".format(type(t))
    )


def _reshape_batch_univariate_transform(t, old_shape, new_shape):
    params = {}
    for name in UNIVARIATE_TRANSFORMS[type(t)]:
        value = getattr(t, name)
        if not isinstance(value, numbers.Number):
            value = value.expand(old_shape).reshape(new_shape)
        params[name] = value
    params["cache_size"] = t._cache_size
    return type(t)(**params)


@reshape_transform_batch.register(torch.distributions.transforms._InverseTransform)
def _(t, old_shape, new_shape):
    return reshape_transform_batch(t.inv, old_shape, new_shape).inv


@reshape_transform_batch.register(dist.transforms.ComposeTransform)
def _(t, old_shape, new_shape):
    return dist.transforms.ComposeTransform(
        [reshape_transform_batch(part, old_shape, new_shape) for part in t.parts]
    )


for _type in UNIVARIATE_TRANSFORMS:
    reshape_transform_batch.register(_type)(_reshape_batch_univariate_transform)


@reshape_transform_batch.register(dist.transforms.HaarTransform)
@reshape_transform_batch.register(dist.transforms.DiscreteCosineTransform)
def _(t, old_shape, new_shape):
    dim = t.event_dim
    assert len(old_shape) >= dim
    assert len(new_shape) >= dim
    if old_shape[-dim:] != new_shape[-dim:]:
        raise NotImplementedError
    return t



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Union, ValuesView

from torch.optim import Optimizer

import pyro

from .optim import PyroOptim


class HorovodOptimizer(PyroOptim):
    r"""
    Distributed wrapper for a :class:`~pyro.optim.optim.PyroOptim` optimizer.

    This class wraps a ``PyroOptim`` object similar to the way
    :func:`horovod.torch.DistributedOptimizer` wraps a
    :class:`torch.optim.Optimizer`.

    .. note::

        This requires :mod:`horovod.torch` to be installed, e.g. via
        ``pip install pyro[horovod]``. For details see
        https://horovod.readthedocs.io/en/stable/install.html

    :param: A Pyro optimizer instance.
    :type pyro_optim: ~pyro.optim.optim.PyroOptim
    :param \*\*horovod_kwargs: Extra parameters passed to
        :func:`horovod.torch.DistributedOptimizer`.
    """

    def __init__(self, pyro_optim: PyroOptim, **horovod_kwargs):
        param_name = pyro.get_param_store().param_name

        def optim_constructor(params, **pt_kwargs) -> Optimizer:
            import horovod.torch as hvd  # type: ignore

            pt_optim = pyro_optim.pt_optim_constructor(params, **pt_kwargs)  # type: ignore
            named_parameters = [(param_name(p), p) for p in params]
            hvd_optim = hvd.DistributedOptimizer(
                pt_optim,
                named_parameters=named_parameters,
                **horovod_kwargs,
            )
            return hvd_optim  # type: ignore

        super().__init__(
            optim_constructor, pyro_optim.pt_optim_args, pyro_optim.pt_clip_args
        )

    def __call__(self, params: Union[List, ValuesView], *args, **kwargs) -> None:
        # Sort by name to ensure deterministic processing order.
        params = sorted(params, key=pyro.get_param_store().param_name)
        super().__call__(params, *args, **kwargs)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn as nn
from torch.distributions import MultivariateNormal, constraints

import pyro.distributions as dist
from pyro.contrib.timeseries.base import TimeSeriesModel
from pyro.nn import PyroParam, pyro_method
from pyro.ops.ssm_gp import MaternKernel
from pyro.ops.tensor_utils import block_diag_embed, repeated_matmul


class GenericLGSSMWithGPNoiseModel(TimeSeriesModel):
    """
    A generic Linear Gaussian State Space Model parameterized with arbitrary time invariant
    transition and observation dynamics together with separate Gaussian Process noise models
    for each output dimension. In more detail, the generative process is:

        :math:`y_i(t) = \\sum_j A_{ij} z_j(t) + f_i(t) + \\epsilon_i(t)`

    where the latent variables :math:`{\\bf z}(t)` follow generic time invariant Linear Gaussian dynamics
    and the :math:`f_i(t)` are Gaussian Processes with Matern kernels.

    The targets are (implicitly) assumed to be evenly spaced in time. In particular a timestep of
    :math:`dt=1.0` for the continuous-time GP dynamics corresponds to a single discrete step of
    the :math:`{\\bf z}`-space dynamics. Training and inference are logarithmic in the length of
    the time series T.

    :param int obs_dim: The dimension of the targets at each time step.
    :param int state_dim: The dimension of the :math:`{\\bf z}` latent state at each time step.
    :param float nu: The order of the Matern kernel; one of 0.5, 1.5 or 2.5.
    :param torch.Tensor length_scale_init: optional initial values for the kernel length scale
        given as a ``obs_dim``-dimensional tensor
    :param torch.Tensor kernel_scale_init: optional initial values for the kernel scale
        given as a ``obs_dim``-dimensional tensor
    :param torch.Tensor obs_noise_scale_init: optional initial values for the observation noise scale
        given as a ``obs_dim``-dimensional tensor
    :param bool learnable_observation_loc: whether the mean of the observation model should be learned or not;
            defaults to False.
    """

    def __init__(
        self,
        obs_dim=1,
        state_dim=2,
        nu=1.5,
        obs_noise_scale_init=None,
        length_scale_init=None,
        kernel_scale_init=None,
        learnable_observation_loc=False,
    ):
        self.obs_dim = obs_dim
        self.state_dim = state_dim
        self.nu = nu

        if obs_noise_scale_init is None:
            obs_noise_scale_init = 0.2 * torch.ones(obs_dim)
        assert obs_noise_scale_init.shape == (obs_dim,)

        super().__init__()

        self.kernel = MaternKernel(
            nu=nu,
            num_gps=obs_dim,
            length_scale_init=length_scale_init,
            kernel_scale_init=kernel_scale_init,
        )
        self.dt = 1.0
        self.full_state_dim = self.kernel.state_dim * obs_dim + state_dim
        self.full_gp_state_dim = self.kernel.state_dim * obs_dim

        self.obs_noise_scale = PyroParam(
            obs_noise_scale_init, constraint=constraints.positive
        )
        self.trans_noise_scale_sq = PyroParam(
            torch.ones(state_dim), constraint=constraints.positive
        )
        self.z_trans_matrix = nn.Parameter(
            torch.eye(state_dim) + 0.03 * torch.randn(state_dim, state_dim)
        )
        self.z_obs_matrix = nn.Parameter(0.3 * torch.randn(state_dim, obs_dim))
        self.init_noise_scale_sq = PyroParam(
            torch.ones(state_dim), constraint=constraints.positive
        )

        gp_obs_matrix = torch.zeros(self.kernel.state_dim * obs_dim, obs_dim)
        for i in range(obs_dim):
            gp_obs_matrix[self.kernel.state_dim * i, i] = 1.0
        self.register_buffer("gp_obs_matrix", gp_obs_matrix)

        self.obs_selector = torch.tensor(
            [self.kernel.state_dim * d for d in range(obs_dim)], dtype=torch.long
        )

        if learnable_observation_loc:
            self.obs_loc = nn.Parameter(torch.zeros(obs_dim))
        else:
            self.register_buffer("obs_loc", torch.zeros(obs_dim))

    def _get_obs_matrix(self):
        # (obs_dim + state_dim, obs_dim) => (gp_state_dim * obs_dim + state_dim, obs_dim)
        return torch.cat([self.gp_obs_matrix, self.z_obs_matrix], dim=0)

    def _get_init_dist(self):
        loc = self.z_trans_matrix.new_zeros(self.full_state_dim)
        covar = self.z_trans_matrix.new_zeros(self.full_state_dim, self.full_state_dim)
        covar[: self.full_gp_state_dim, : self.full_gp_state_dim] = block_diag_embed(
            self.kernel.stationary_covariance()
        )
        covar[self.full_gp_state_dim :, self.full_gp_state_dim :] = (
            self.init_noise_scale_sq.diag_embed()
        )
        return MultivariateNormal(loc, covar)

    def _get_obs_dist(self):
        return dist.Normal(self.obs_loc, self.obs_noise_scale).to_event(1)

    def get_dist(self, duration=None):
        """
        Get the :class:`~pyro.distributions.GaussianHMM` distribution that corresponds
        to :class:`GenericLGSSMWithGPNoiseModel`.

        :param int duration: Optional size of the time axis ``event_shape[0]``.
            This is required when sampling from homogeneous HMMs whose parameters
            are not expanded along the time axis.
        """
        (
            gp_trans_matrix,
            gp_process_covar,
        ) = self.kernel.transition_matrix_and_covariance(dt=self.dt)

        trans_covar = self.z_trans_matrix.new_zeros(
            self.full_state_dim, self.full_state_dim
        )
        trans_covar[: self.full_gp_state_dim, : self.full_gp_state_dim] = (
            block_diag_embed(gp_process_covar)
        )
        trans_covar[self.full_gp_state_dim :, self.full_gp_state_dim :] = (
            self.trans_noise_scale_sq.diag_embed()
        )
        trans_dist = MultivariateNormal(
            trans_covar.new_zeros(self.full_state_dim), trans_covar
        )

        full_trans_mat = trans_covar.new_zeros(self.full_state_dim, self.full_state_dim)
        full_trans_mat[: self.full_gp_state_dim, : self.full_gp_state_dim] = (
            block_diag_embed(gp_trans_matrix)
        )
        full_trans_mat[self.full_gp_state_dim :, self.full_gp_state_dim :] = (
            self.z_trans_matrix
        )

        return dist.GaussianHMM(
            self._get_init_dist(),
            full_trans_mat,
            trans_dist,
            self._get_obs_matrix(),
            self._get_obs_dist(),
            duration=duration,
        )

    @pyro_method
    def log_prob(self, targets):
        """
        :param torch.Tensor targets: A 2-dimensional tensor of real-valued targets
            of shape ``(T, obs_dim)``, where ``T`` is the length of the time series and ``obs_dim``
            is the dimension of the real-valued ``targets`` at each time step
        :returns torch.Tensor: A (scalar) log probability.
        """
        assert targets.dim() == 2 and targets.size(-1) == self.obs_dim
        return self.get_dist().log_prob(targets)

    @torch.no_grad()
    def _filter(self, targets):
        """
        Return the filtering state for the associated state space model.
        """
        assert targets.dim() == 2 and targets.size(-1) == self.obs_dim
        return self.get_dist().filter(targets)

    @torch.no_grad()
    def _forecast(self, N_timesteps, filtering_state, include_observation_noise=True):
        """
        Internal helper for forecasting.
        """
        dts = (
            torch.arange(
                N_timesteps,
                dtype=self.z_trans_matrix.dtype,
                device=self.z_trans_matrix.device,
            )
            + 1.0
        )
        dts = dts.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)

        (
            gp_trans_matrix,
            gp_process_covar,
        ) = self.kernel.transition_matrix_and_covariance(dt=dts)
        gp_trans_matrix = block_diag_embed(gp_trans_matrix)
        gp_process_covar = block_diag_embed(gp_process_covar[..., 0:1, 0:1])

        N_trans_matrix = repeated_matmul(self.z_trans_matrix, N_timesteps)
        N_trans_obs = torch.matmul(N_trans_matrix, self.z_obs_matrix)

        # z-state contribution + gp contribution
        predicted_mean1 = torch.matmul(
            filtering_state.loc[-self.state_dim :].unsqueeze(-2), N_trans_obs
        ).squeeze(-2)
        predicted_mean2 = torch.matmul(
            filtering_state.loc[: self.full_gp_state_dim].unsqueeze(-2),
            gp_trans_matrix[..., self.obs_selector],
        ).squeeze(-2)
        predicted_mean = predicted_mean1 + predicted_mean2

        # first compute the contributions from filtering_state.covariance_matrix: z-space and gp
        fs_cov = filtering_state.covariance_matrix
        predicted_covar1z = torch.matmul(
            N_trans_obs.transpose(-1, -2),
            torch.matmul(
                fs_cov[self.full_gp_state_dim :, self.full_gp_state_dim :], N_trans_obs
            ),
        )  # N O O
        gp_trans = gp_trans_matrix[..., self.obs_selector]
        predicted_covar1gp = torch.matmul(
            gp_trans.transpose(-1, -2),
            torch.matmul(
                fs_cov[: self.full_gp_state_dim :, : self.full_gp_state_dim], gp_trans
            ),
        )

        # next compute the contribution from process noise that is injected at each timestep.
        # (we need to do a cumulative sum to integrate across time for the z-state contribution)
        z_process_covar = self.trans_noise_scale_sq.diag_embed()
        N_trans_obs_shift = torch.cat(
            [self.z_obs_matrix.unsqueeze(0), N_trans_obs[0:-1]]
        )
        predicted_covar2z = torch.matmul(
            N_trans_obs_shift.transpose(-1, -2),
            torch.matmul(z_process_covar, N_trans_obs_shift),
        )  # N O O

        predicted_covar = (
            predicted_covar1z
            + predicted_covar1gp
            + gp_process_covar
            + torch.cumsum(predicted_covar2z, dim=0)
        )

        if include_observation_noise:
            predicted_covar = (
                predicted_covar + self.obs_noise_scale.pow(2.0).diag_embed()
            )

        return predicted_mean, predicted_covar

    @pyro_method
    def forecast(self, targets, N_timesteps):
        """
        :param torch.Tensor targets: A 2-dimensional tensor of real-valued targets
            of shape ``(T, obs_dim)``, where ``T`` is the length of the time series and ``obs_dim``
            is the dimension of the real-valued targets at each time step. These
            represent the training data that are conditioned on for the purpose of making
            forecasts.
        :param int N_timesteps: The number of timesteps to forecast into the future from
            the final target ``targets[-1]``.
        :returns torch.distributions.MultivariateNormal: Returns a predictive MultivariateNormal distribution
            with batch shape ``(N_timesteps,)`` and event shape ``(obs_dim,)``
        """
        filtering_state = self._filter(targets)
        predicted_mean, predicted_covar = self._forecast(N_timesteps, filtering_state)
        return MultivariateNormal(predicted_mean, predicted_covar)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math

import torch
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all

from pyro.distributions.torch_distribution import TorchDistribution


def _unsafe_standard_stable(alpha, beta, V, W, coords):
    # Implements a noisily reparametrized version of the sampler
    # Chambers-Mallows-Stuck method as corrected by Weron [1,3] and simplified
    # by Nolan [4]. This will fail if alpha is close to 1.

    # Differentiably transform noise via parameters.
    assert V.shape == W.shape
    inv_alpha = alpha.reciprocal()
    half_pi = math.pi / 2
    eps = torch.finfo(V.dtype).eps
    # make V belong to the open interval (-pi/2, pi/2)
    V = V.clamp(min=2 * eps - half_pi, max=half_pi - 2 * eps)
    ha = half_pi * alpha
    b = beta * ha.tan()
    # +/- `ha` term to keep the precision of alpha * (V + half_pi) when V ~ -half_pi
    v = b.atan() - ha + alpha * (V + half_pi)
    Z = (
        v.sin()
        / ((1 + b * b).rsqrt() * V.cos()).pow(inv_alpha)
        * ((v - V).cos().clamp(min=eps) / W).pow(inv_alpha - 1)
    )
    Z.data[Z.data != Z.data] = 0  # drop occasional NANs

    # Optionally convert to Nolan's parametrization S^0 where samples depend
    # continuously on (alpha,beta), allowing interpolation around the hole at
    # alpha=1.
    if coords == "S0":
        return Z - b
    elif coords == "S":
        return Z
    else:
        raise ValueError("Unknown coords: {}".format(coords))


RADIUS = 0.01


def _standard_stable(alpha, beta, aux_uniform, aux_exponential, coords):
    """
    Differentiably transform two random variables::

        aux_uniform ~ Uniform(-pi/2, pi/2)
        aux_exponential ~ Exponential(1)

    to a standard ``Stable(alpha, beta)`` random variable.
    """
    # Determine whether a hole workaround is needed.
    with torch.no_grad():
        hole = 1.0
        near_hole = (alpha - hole).abs() <= RADIUS
    if not torch._C._get_tracing_state() and not near_hole.any():
        return _unsafe_standard_stable(
            alpha, beta, aux_uniform, aux_exponential, coords=coords
        )
    if coords == "S":
        # S coords are discontinuous, so interpolate instead in S0 coords.
        Z = _standard_stable(alpha, beta, aux_uniform, aux_exponential, "S0")
        return torch.where(alpha == 1, Z, Z + beta * (math.pi / 2 * alpha).tan())

    # Avoid the hole at alpha=1 by interpolating between pairs
    # of points at hole-RADIUS and hole+RADIUS.
    aux_uniform_ = aux_uniform.unsqueeze(-1)
    aux_exponential_ = aux_exponential.unsqueeze(-1)
    beta_ = beta.unsqueeze(-1)
    alpha_ = alpha.unsqueeze(-1).expand(alpha.shape + (2,)).contiguous()
    with torch.no_grad():
        lower, upper = alpha_.unbind(-1)
        lower.data[near_hole] = hole - RADIUS
        upper.data[near_hole] = hole + RADIUS
        # We don't need to backprop through weights, since we've pretended
        # alpha_ is reparametrized, even though we've clamped some values.
        #               |a - a'|
        # weight = 1 - ----------
        #              2 * RADIUS
        weights = (alpha_ - alpha.unsqueeze(-1)).abs_().mul_(-1 / (2 * RADIUS)).add_(1)
        weights[~near_hole] = 0.5
    pairs = _unsafe_standard_stable(
        alpha_, beta_, aux_uniform_, aux_exponential_, coords=coords
    )
    return (pairs * weights).sum(-1)


class Stable(TorchDistribution):
    r"""
    Levy :math:`\alpha`-stable distribution. See [1] for a review.

    This uses Nolan's parametrization [2] of the ``loc`` parameter, which is
    required for continuity and differentiability. This corresponds to the
    notation :math:`S^0_\alpha(\beta,\sigma,\mu_0)` of [1], where
    :math:`\alpha` = stability, :math:`\beta` = skew, :math:`\sigma` = scale,
    and :math:`\mu_0` = loc. To instead use the S parameterization as in scipy,
    pass ``coords="S"``, but BEWARE this is discontinuous at ``stability=1``
    and has poor geometry for inference.

    This implements a reparametrized sampler :meth:`rsample` , but does not
    implement :meth:`log_prob` . Inference can be performed using either
    likelihood-free algorithms such as
    :class:`~pyro.infer.energy_distance.EnergyDistance`, or reparameterization
    via the :func:`~pyro.poutine.handlers.reparam` handler with one of the
    reparameterizers :class:`~pyro.infer.reparam.stable.LatentStableReparam` ,
    :class:`~pyro.infer.reparam.stable.SymmetricStableReparam` , or
    :class:`~pyro.infer.reparam.stable.StableReparam` e.g.::

        with poutine.reparam(config={"x": StableReparam()}):
            pyro.sample("x", Stable(stability, skew, scale, loc))

    or simply wrap in :class:`~pyro.infer.reparam.strategies.MinimalReparam` or
    :class:`~pyro.infer.reparam.strategies.AutoReparam` , e.g.::

        @MinimalReparam()
        def model():
            ...

    [1] S. Borak, W. Hardle, R. Weron (2005).
        Stable distributions.
        https://edoc.hu-berlin.de/bitstream/handle/18452/4526/8.pdf
    [2] J.P. Nolan (1997).
        Numerical calculation of stable densities and distribution functions.
    [3] Rafal Weron (1996).
        On the Chambers-Mallows-Stuck Method for
        Simulating Skewed Stable Random Variables.
    [4] J.P. Nolan (2017).
        Stable Distributions: Models for Heavy Tailed Data.
        https://edspace.american.edu/jpnolan/wp-content/uploads/sites/1720/2020/09/Chap1.pdf

    :param Tensor stability: Levy stability parameter :math:`\alpha\in(0,2]` .
    :param Tensor skew: Skewness :math:`\beta\in[-1,1]` .
    :param Tensor scale: Scale :math:`\sigma > 0` . Defaults to 1.
    :param Tensor loc: Location :math:`\mu_0` when using Nolan's S0
        parametrization [2], or :math:`\mu` when using the S parameterization.
        Defaults to 0.
    :param str coords: Either "S0" (default) to use Nolan's continuous S0
        parametrization, or "S" to use the discontinuous parameterization.
    """

    has_rsample = True
    arg_constraints = {
        "stability": constraints.interval(0, 2),  # half-open (0, 2]
        "skew": constraints.interval(-1, 1),  # closed [-1, 1]
        "scale": constraints.positive,
        "loc": constraints.real,
    }
    support = constraints.real

    def __init__(
        self, stability, skew, scale=1.0, loc=0.0, coords="S0", validate_args=None
    ):
        assert coords in ("S", "S0"), coords
        self.stability, self.skew, self.scale, self.loc = broadcast_all(
            stability, skew, scale, loc
        )
        self.coords = coords
        super().__init__(self.loc.shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(Stable, _instance)
        batch_shape = torch.Size(batch_shape)
        for name in self.arg_constraints:
            setattr(new, name, getattr(self, name).expand(batch_shape))
        new.coords = self.coords
        super(Stable, new).__init__(batch_shape, validate_args=False)
        new._validate_args = self._validate_args
        return new

    def log_prob(self, value):
        raise NotImplementedError("Stable.log_prob() is not implemented")

    def rsample(self, sample_shape=torch.Size()):
        # Draw parameter-free noise.
        with torch.no_grad():
            shape = self._extended_shape(sample_shape)
            new_empty = self.stability.new_empty
            aux_uniform = new_empty(shape).uniform_(-math.pi / 2, math.pi / 2)
            aux_exponential = new_empty(shape).exponential_()

        # Differentiably transform.
        x = _standard_stable(
            self.stability, self.skew, aux_uniform, aux_exponential, coords=self.coords
        )
        return self.loc + self.scale * x

    @property
    def mean(self):
        result = self.loc
        if self.coords == "S0":
            result = (
                result - self.scale * self.skew * (math.pi / 2 * self.stability).tan()
            )
        return result.masked_fill(self.stability <= 1, math.nan)

    @property
    def variance(self):
        var = self.scale * self.scale
        return var.mul(2).masked_fill(self.stability < 2, math.inf)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

"""
Example MuE observation models.
"""

import datetime

import numpy as np
import torch
import torch.nn as nn
from torch.nn.functional import softplus
from torch.optim import Adam
from torch.utils.data import DataLoader

import pyro
import pyro.distributions as dist
from pyro import poutine
from pyro.contrib.mue.missingdatahmm import MissingDataDiscreteHMM
from pyro.contrib.mue.statearrangers import Profile
from pyro.infer import SVI, JitTrace_ELBO, Trace_ELBO
from pyro.optim import MultiStepLR


class ProfileHMM(nn.Module):
    """
    Profile HMM.

    This model consists of a constant distribution (a delta function) over the
    regressor sequence, plus a MuE observation distribution. The priors
    are all Normal distributions, and are pushed through a softmax function
    onto the simplex.

    :param int latent_seq_length: Length of the latent regressor sequence M.
        Must be greater than or equal to 1.
    :param int alphabet_length: Length of the sequence alphabet (e.g. 20 for
        amino acids).
    :param float prior_scale: Standard deviation of the prior distribution.
    :param float indel_prior_bias: Mean of the prior distribution over the
        log probability of an indel not occurring. Higher values lead to lower
        probability of indels.
    :param bool cuda: Transfer data onto the GPU during training.
    :param bool pin_memory: Pin memory for faster GPU transfer.
    """

    def __init__(
        self,
        latent_seq_length,
        alphabet_length,
        prior_scale=1.0,
        indel_prior_bias=10.0,
        cuda=False,
        pin_memory=False,
    ):
        super().__init__()
        assert isinstance(cuda, bool)
        self.is_cuda = cuda
        assert isinstance(pin_memory, bool)
        self.pin_memory = pin_memory

        assert isinstance(latent_seq_length, int) and latent_seq_length > 0
        self.latent_seq_length = latent_seq_length
        assert isinstance(alphabet_length, int) and alphabet_length > 0
        self.alphabet_length = alphabet_length

        self.precursor_seq_shape = (latent_seq_length, alphabet_length)
        self.insert_seq_shape = (latent_seq_length + 1, alphabet_length)
        self.indel_shape = (latent_seq_length, 3, 2)

        assert isinstance(prior_scale, float)
        self.prior_scale = prior_scale
        assert isinstance(indel_prior_bias, float)
        self.indel_prior = torch.tensor([indel_prior_bias, 0.0])

        # Initialize state arranger.
        self.statearrange = Profile(latent_seq_length)

    def model(self, seq_data, local_scale):
        # Latent sequence.
        precursor_seq = pyro.sample(
            "precursor_seq",
            dist.Normal(
                torch.zeros(self.precursor_seq_shape),
                self.prior_scale * torch.ones(self.precursor_seq_shape),
            ).to_event(2),
        )
        precursor_seq_logits = precursor_seq - precursor_seq.logsumexp(-1, True)
        insert_seq = pyro.sample(
            "insert_seq",
            dist.Normal(
                torch.zeros(self.insert_seq_shape),
                self.prior_scale * torch.ones(self.insert_seq_shape),
            ).to_event(2),
        )
        insert_seq_logits = insert_seq - insert_seq.logsumexp(-1, True)

        # Indel probabilities.
        insert = pyro.sample(
            "insert",
            dist.Normal(
                self.indel_prior * torch.ones(self.indel_shape),
                self.prior_scale * torch.ones(self.indel_shape),
            ).to_event(3),
        )
        insert_logits = insert - insert.logsumexp(-1, True)
        delete = pyro.sample(
            "delete",
            dist.Normal(
                self.indel_prior * torch.ones(self.indel_shape),
                self.prior_scale * torch.ones(self.indel_shape),
            ).to_event(3),
        )
        delete_logits = delete - delete.logsumexp(-1, True)

        # Construct HMM parameters.
        initial_logits, transition_logits, observation_logits = self.statearrange(
            precursor_seq_logits, insert_seq_logits, insert_logits, delete_logits
        )

        with pyro.plate("batch", seq_data.shape[0]):
            with poutine.scale(scale=local_scale):
                # Observations.
                pyro.sample(
                    "obs_seq",
                    MissingDataDiscreteHMM(
                        initial_logits, transition_logits, observation_logits
                    ),
                    obs=seq_data,
                )

    def guide(self, seq_data, local_scale):
        # Sequence.
        precursor_seq_q_mn = pyro.param(
            "precursor_seq_q_mn", torch.zeros(self.precursor_seq_shape)
        )
        precursor_seq_q_sd = pyro.param(
            "precursor_seq_q_sd", torch.zeros(self.precursor_seq_shape)
        )
        pyro.sample(
            "precursor_seq",
            dist.Normal(precursor_seq_q_mn, softplus(precursor_seq_q_sd)).to_event(2),
        )
        insert_seq_q_mn = pyro.param(
            "insert_seq_q_mn", torch.zeros(self.insert_seq_shape)
        )
        insert_seq_q_sd = pyro.param(
            "insert_seq_q_sd", torch.zeros(self.insert_seq_shape)
        )
        pyro.sample(
            "insert_seq",
            dist.Normal(insert_seq_q_mn, softplus(insert_seq_q_sd)).to_event(2),
        )

        # Indels.
        insert_q_mn = pyro.param(
            "insert_q_mn", torch.ones(self.indel_shape) * self.indel_prior
        )
        insert_q_sd = pyro.param("insert_q_sd", torch.zeros(self.indel_shape))
        pyro.sample(
            "insert",
            dist.Normal(insert_q_mn, softplus(insert_q_sd)).to_event(3),
        )
        delete_q_mn = pyro.param(
            "delete_q_mn", torch.ones(self.indel_shape) * self.indel_prior
        )
        delete_q_sd = pyro.param("delete_q_sd", torch.zeros(self.indel_shape))
        pyro.sample(
            "delete",
            dist.Normal(delete_q_mn, softplus(delete_q_sd)).to_event(3),
        )

    def fit_svi(
        self,
        dataset,
        epochs=2,
        batch_size=1,
        scheduler=None,
        jit=False,
    ):
        """
        Infer approximate posterior with stochastic variational inference.

        This runs :class:`~pyro.infer.svi.SVI`. It is an approximate inference
        method useful for quickly iterating on probabilistic models.

        :param ~torch.utils.data.Dataset dataset: The training dataset.
        :param int epochs: Number of epochs of training.
        :param int batch_size: Minibatch size (number of sequences).
        :param pyro.optim.MultiStepLR scheduler: Optimization scheduler.
            (Default: Adam optimizer, 0.01 constant learning rate.)
        :param bool jit: Whether to use a jit compiled ELBO.
        """

        # Setup.
        if batch_size is not None:
            self.batch_size = batch_size
        if scheduler is None:
            scheduler = MultiStepLR(
                {
                    "optimizer": Adam,
                    "optim_args": {"lr": 0.01},
                    "milestones": [],
                    "gamma": 0.5,
                }
            )
        if self.is_cuda:
            device = torch.device("cuda")
        else:
            device = torch.device("cpu")
        # Initialize guide.
        self.guide(None, None)
        dataload = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=self.pin_memory,
            generator=torch.Generator(device=device),
        )
        # Setup stochastic variational inference.
        if jit:
            elbo = JitTrace_ELBO(ignore_jit_warnings=True)
        else:
            elbo = Trace_ELBO()
        svi = SVI(self.model, self.guide, scheduler, loss=elbo)

        # Run inference.
        losses = []
        t0 = datetime.datetime.now()
        for epoch in range(epochs):
            for seq_data, L_data in dataload:
                if self.is_cuda:
                    seq_data = seq_data.cuda()
                loss = svi.step(
                    seq_data, torch.tensor(len(dataset) / seq_data.shape[0])
                )
                losses.append(loss)
                scheduler.step()
            print(epoch, loss, " ", datetime.datetime.now() - t0)
        return losses

    def evaluate(self, dataset_train, dataset_test=None, jit=False):
        """
        Evaluate performance (log probability and per residue perplexity) on
        train and test datasets.

        :param ~torch.utils.data.Dataset dataset: The training dataset.
        :param ~torch.utils.data.Dataset dataset: The testing dataset.
        :param bool jit: Whether to use a jit compiled ELBO.
        """
        dataload_train = DataLoader(dataset_train, batch_size=1, shuffle=False)
        if dataset_test is not None:
            dataload_test = DataLoader(dataset_test, batch_size=1, shuffle=False)
        # Initialize guide.
        self.guide(None, None)
        if jit:
            elbo = JitTrace_ELBO(ignore_jit_warnings=True)
        else:
            elbo = Trace_ELBO()
        scheduler = MultiStepLR({"optimizer": Adam, "optim_args": {"lr": 0.01}})
        # Setup stochastic variational inference.
        svi = SVI(self.model, self.guide, scheduler, loss=elbo)

        # Compute elbo and perplexity.
        train_lp, train_perplex = self._evaluate_local_elbo(
            svi, dataload_train, len(dataset_train)
        )
        if dataset_test is not None:
            test_lp, test_perplex = self._evaluate_local_elbo(
                svi, dataload_test, len(dataset_test)
            )
            return train_lp, test_lp, train_perplex, test_perplex
        else:
            return train_lp, None, train_perplex, None

    def _local_variables(self, name, site):
        """Return per datapoint random variables in model."""
        return name in ["obs_L", "obs_seq"]

    def _evaluate_local_elbo(self, svi, dataload, data_size):
        """Evaluate elbo and average per residue perplexity."""
        lp, perplex = 0.0, 0.0
        with torch.no_grad():
            for seq_data, L_data in dataload:
                if self.is_cuda:
                    seq_data, L_data = seq_data.cuda(), L_data.cuda()
                conditioned_model = poutine.condition(
                    self.model, data={"obs_seq": seq_data}
                )
                args = (seq_data, torch.tensor(1.0))
                guide_tr = poutine.trace(self.guide).get_trace(*args)
                model_tr = poutine.trace(
                    poutine.replay(conditioned_model, trace=guide_tr)
                ).get_trace(*args)
                local_elbo = (
                    (
                        model_tr.log_prob_sum(self._local_variables)
                        - guide_tr.log_prob_sum(self._local_variables)
                    )
                    .cpu()
                    .numpy()
                )
                lp += local_elbo
                perplex += -local_elbo / L_data[0].cpu().numpy()
        perplex = np.exp(perplex / data_size)
        return lp, perplex


class Encoder(nn.Module):
    def __init__(self, data_length, alphabet_length, z_dim):
        super().__init__()

        self.input_size = data_length * alphabet_length
        self.f1_mn = nn.Linear(self.input_size, z_dim)
        self.f1_sd = nn.Linear(self.input_size, z_dim)

    def forward(self, data):
        data = data.reshape(-1, self.input_size)
        z_loc = self.f1_mn(data)
        z_scale = softplus(self.f1_sd(data))

        return z_loc, z_scale


class FactorMuE(nn.Module):
    """
    FactorMuE

    This model consists of probabilistic PCA plus a MuE output distribution.

    The priors are all Normal distributions, and where relevant pushed through
    a softmax onto the simplex.

    :param int data_length: Length of the input sequence matrix, including
        zero padding at the end.
    :param int alphabet_length: Length of the sequence alphabet (e.g. 20 for
        amino acids).
    :param int z_dim: Number of dimensions of the z space.
    :param int batch_size: Minibatch size.
    :param int latent_seq_length: Length of the latent regressor sequence (M).
        Must be greater than or equal to 1. (Default: 1.1 x data_length.)
    :param bool indel_factor_dependence: Indel probabilities depend on the
        latent variable z.
    :param float indel_prior_scale: Standard deviation of the prior
        distribution on indel parameters.
    :param float indel_prior_bias: Mean of the prior distribution over the
        log probability of an indel not occurring. Higher values lead to lower
        probability of indels.
    :param float inverse_temp_prior: Mean of the prior distribution over the
        inverse temperature parameter.
    :param float weights_prior_scale: Standard deviation of the prior
        distribution over the factors.
    :param float offset_prior_scale: Standard deviation of the prior
        distribution over the offset (constant) in the pPCA model.
    :param str z_prior_distribution: Prior distribution over the latent
        variable z. Either 'Normal' (pPCA model) or 'Laplace' (an ICA model).
    :param bool ARD_prior: Use automatic relevance determination prior on
        factors.
    :param bool substitution_matrix: Use a learnable substitution matrix
        rather than the identity matrix.
    :param float substitution_prior_scale: Standard deviation of the prior
        distribution over substitution matrix parameters (when
        substitution_matrix is True).
    :param int latent_alphabet_length: Length of the alphabet in the latent
        regressor sequence.
    :param bool cuda: Transfer data onto the GPU during training.
    :param bool pin_memory: Pin memory for faster GPU transfer.
    :param float epsilon: A small value for numerical stability.
    """

    def __init__(
        self,
        data_length,
        alphabet_length,
        z_dim,
        batch_size=10,
        latent_seq_length=None,
        indel_factor_dependence=False,
        indel_prior_scale=1.0,
        indel_prior_bias=10.0,
        inverse_temp_prior=100.0,
        weights_prior_scale=1.0,
        offset_prior_scale=1.0,
        z_prior_distribution="Normal",
        ARD_prior=False,
        substitution_matrix=True,
        substitution_prior_scale=10.0,
        latent_alphabet_length=None,
        cuda=False,
        pin_memory=False,
        epsilon=1e-32,
    ):
        super().__init__()
        assert isinstance(cuda, bool)
        self.is_cuda = cuda
        assert isinstance(pin_memory, bool)
        self.pin_memory = pin_memory

        # Constants.
        assert isinstance(data_length, int) and data_length > 0
        self.data_length = data_length
        if latent_seq_length is None:
            latent_seq_length = int(data_length * 1.1)
        else:
            assert isinstance(latent_seq_length, int) and latent_seq_length > 0
        self.latent_seq_length = latent_seq_length
        assert isinstance(alphabet_length, int) and alphabet_length > 0
        self.alphabet_length = alphabet_length
        assert isinstance(z_dim, int) and z_dim > 0
        self.z_dim = z_dim

        # Parameter shapes.
        if (not substitution_matrix) or (latent_alphabet_length is None):
            latent_alphabet_length = alphabet_length
        self.latent_alphabet_length = latent_alphabet_length
        self.indel_shape = (latent_seq_length, 3, 2)
        self.total_factor_size = (
            (2 * latent_seq_length + 1) * latent_alphabet_length
            + 2 * indel_factor_dependence * latent_seq_length * 3 * 2
        )

        # Architecture.
        self.indel_factor_dependence = indel_factor_dependence
        self.ARD_prior = ARD_prior
        self.substitution_matrix = substitution_matrix

        # Priors.
        assert isinstance(indel_prior_scale, float)
        self.indel_prior_scale = torch.tensor(indel_prior_scale)
        assert isinstance(indel_prior_bias, float)
        self.indel_prior = torch.tensor([indel_prior_bias, 0.0])
        assert isinstance(inverse_temp_prior, float)
        self.inverse_temp_prior = torch.tensor(inverse_temp_prior)
        assert isinstance(weights_prior_scale, float)
        self.weights_prior_scale = torch.tensor(weights_prior_scale)
        assert isinstance(offset_prior_scale, float)
        self.offset_prior_scale = torch.tensor(offset_prior_scale)
        assert isinstance(epsilon, float)
        self.epsilon = torch.tensor(epsilon)
        assert isinstance(substitution_prior_scale, float)
        self.substitution_prior_scale = torch.tensor(substitution_prior_scale)
        self.z_prior_distribution = z_prior_distribution

        # Batch control.
        assert isinstance(batch_size, int)
        self.batch_size = batch_size

        # Initialize layers.
        self.encoder = Encoder(data_length, alphabet_length, z_dim)
        self.statearrange = Profile(latent_seq_length)

    def decoder(self, z, W, B, inverse_temp):
        # Project.
        v = torch.mm(z, W) + B

        out = dict()
        if self.indel_factor_dependence:
            # Extract insertion and deletion parameters.
            ind0 = (2 * self.latent_seq_length + 1) * self.latent_alphabet_length
            ind1 = ind0 + self.latent_seq_length * 3 * 2
            ind2 = ind1 + self.latent_seq_length * 3 * 2
            insert_v, delete_v = v[:, ind0:ind1], v[:, ind1:ind2]
            insert_v = (
                insert_v.reshape([-1, self.latent_seq_length, 3, 2]) + self.indel_prior
            )
            out["insert_logits"] = insert_v - insert_v.logsumexp(-1, True)
            delete_v = (
                delete_v.reshape([-1, self.latent_seq_length, 3, 2]) + self.indel_prior
            )
            out["delete_logits"] = delete_v - delete_v.logsumexp(-1, True)
        # Extract precursor and insertion sequences.
        ind0 = self.latent_seq_length * self.latent_alphabet_length
        ind1 = ind0 + (self.latent_seq_length + 1) * self.latent_alphabet_length
        precursor_seq_v, insert_seq_v = v[:, :ind0], v[:, ind0:ind1]
        precursor_seq_v = (precursor_seq_v * softplus(inverse_temp)).reshape(
            [-1, self.latent_seq_length, self.latent_alphabet_length]
        )
        out["precursor_seq_logits"] = precursor_seq_v - precursor_seq_v.logsumexp(
            -1, True
        )
        insert_seq_v = (insert_seq_v * softplus(inverse_temp)).reshape(
            [-1, self.latent_seq_length + 1, self.latent_alphabet_length]
        )
        out["insert_seq_logits"] = insert_seq_v - insert_seq_v.logsumexp(-1, True)

        return out

    def model(self, seq_data, local_scale, local_prior_scale):
        # ARD prior.
        if self.ARD_prior:
            # Relevance factors
            alpha = pyro.sample(
                "alpha",
                dist.Gamma(torch.ones(self.z_dim), torch.ones(self.z_dim)).to_event(1),
            )
        else:
            alpha = torch.ones(self.z_dim)

        # Factor and offset.
        W = pyro.sample(
            "W",
            dist.Normal(
                torch.zeros([self.z_dim, self.total_factor_size]),
                torch.ones([self.z_dim, self.total_factor_size])
                * self.weights_prior_scale
                / (alpha[:, None] + self.epsilon),
            ).to_event(2),
        )
        B = pyro.sample(
            "B",
            dist.Normal(
                torch.zeros(self.total_factor_size),
                torch.ones(self.total_factor_size) * self.offset_prior_scale,
            ).to_event(1),
        )

        # Indel probabilities.
        if not self.indel_factor_dependence:
            insert = pyro.sample(
                "insert",
                dist.Normal(
                    self.indel_prior * torch.ones(self.indel_shape),
                    self.indel_prior_scale * torch.ones(self.indel_shape),
                ).to_event(3),
            )
            insert_logits = insert - insert.logsumexp(-1, True)
            delete = pyro.sample(
                "delete",
                dist.Normal(
                    self.indel_prior * torch.ones(self.indel_shape),
                    self.indel_prior_scale * torch.ones(self.indel_shape),
                ).to_event(3),
            )
            delete_logits = delete - delete.logsumexp(-1, True)

        # Inverse temperature.
        inverse_temp = pyro.sample(
            "inverse_temp", dist.Normal(self.inverse_temp_prior, torch.tensor(1.0))
        )

        # Substitution matrix.
        if self.substitution_matrix:
            substitute = pyro.sample(
                "substitute",
                dist.Normal(
                    torch.zeros([self.latent_alphabet_length, self.alphabet_length]),
                    self.substitution_prior_scale
                    * torch.ones([self.latent_alphabet_length, self.alphabet_length]),
                ).to_event(2),
            )

        with pyro.plate("batch", seq_data.shape[0]):
            with poutine.scale(scale=local_scale):
                with poutine.scale(scale=local_prior_scale):
                    # Sample latent variable from prior.
                    if self.z_prior_distribution == "Normal":
                        z = pyro.sample(
                            "latent",
                            dist.Normal(
                                torch.zeros(self.z_dim), torch.ones(self.z_dim)
                            ).to_event(1),
                        )
                    elif self.z_prior_distribution == "Laplace":
                        z = pyro.sample(
                            "latent",
                            dist.Laplace(
                                torch.zeros(self.z_dim), torch.ones(self.z_dim)
                            ).to_event(1),
                        )

                # Decode latent sequence.
                decoded = self.decoder(z, W, B, inverse_temp)
                if self.indel_factor_dependence:
                    insert_logits = decoded["insert_logits"]
                    delete_logits = decoded["delete_logits"]

                # Construct HMM parameters.
                if self.substitution_matrix:
                    (
                        initial_logits,
                        transition_logits,
                        observation_logits,
                    ) = self.statearrange(
                        decoded["precursor_seq_logits"],
                        decoded["insert_seq_logits"],
                        insert_logits,
                        delete_logits,
                        substitute,
                    )
                else:
                    (
                        initial_logits,
                        transition_logits,
                        observation_logits,
                    ) = self.statearrange(
                        decoded["precursor_seq_logits"],
                        decoded["insert_seq_logits"],
                        insert_logits,
                        delete_logits,
                    )
                # Draw samples.
                pyro.sample(
                    "obs_seq",
                    MissingDataDiscreteHMM(
                        initial_logits, transition_logits, observation_logits
                    ),
                    obs=seq_data,
                )

    def guide(self, seq_data, local_scale, local_prior_scale):
        # Register encoder with pyro.
        pyro.module("encoder", self.encoder)

        # ARD weightings.
        if self.ARD_prior:
            alpha_conc = pyro.param("alpha_conc", torch.randn(self.z_dim))
            alpha_rate = pyro.param("alpha_rate", torch.randn(self.z_dim))
            pyro.sample(
                "alpha",
                dist.Gamma(softplus(alpha_conc), softplus(alpha_rate)).to_event(1),
            )
        # Factors.
        W_q_mn = pyro.param("W_q_mn", torch.randn([self.z_dim, self.total_factor_size]))
        W_q_sd = pyro.param("W_q_sd", torch.ones([self.z_dim, self.total_factor_size]))
        pyro.sample("W", dist.Normal(W_q_mn, softplus(W_q_sd)).to_event(2))
        B_q_mn = pyro.param("B_q_mn", torch.randn(self.total_factor_size))
        B_q_sd = pyro.param("B_q_sd", torch.ones(self.total_factor_size))
        pyro.sample("B", dist.Normal(B_q_mn, softplus(B_q_sd)).to_event(1))

        # Indel probabilities.
        if not self.indel_factor_dependence:
            insert_q_mn = pyro.param(
                "insert_q_mn", torch.ones(self.indel_shape) * self.indel_prior
            )
            insert_q_sd = pyro.param("insert_q_sd", torch.zeros(self.indel_shape))
            pyro.sample(
                "insert", dist.Normal(insert_q_mn, softplus(insert_q_sd)).to_event(3)
            )
            delete_q_mn = pyro.param(
                "delete_q_mn", torch.ones(self.indel_shape) * self.indel_prior
            )
            delete_q_sd = pyro.param("delete_q_sd", torch.zeros(self.indel_shape))
            pyro.sample(
                "delete", dist.Normal(delete_q_mn, softplus(delete_q_sd)).to_event(3)
            )

        # Inverse temperature.
        inverse_temp_q_mn = pyro.param("inverse_temp_q_mn", torch.tensor(0.0))
        inverse_temp_q_sd = pyro.param("inverse_temp_q_sd", torch.tensor(0.0))
        pyro.sample(
            "inverse_temp", dist.Normal(inverse_temp_q_mn, softplus(inverse_temp_q_sd))
        )

        # Substitution matrix.
        if self.substitution_matrix:
            substitute_q_mn = pyro.param(
                "substitute_q_mn",
                torch.zeros([self.latent_alphabet_length, self.alphabet_length]),
            )
            substitute_q_sd = pyro.param(
                "substitute_q_sd",
                torch.zeros([self.latent_alphabet_length, self.alphabet_length]),
            )
            pyro.sample(
                "substitute",
                dist.Normal(substitute_q_mn, softplus(substitute_q_sd)).to_event(2),
            )

        # Per datapoint local latent variables.
        with pyro.plate("batch", seq_data.shape[0]):
            # Encode sequences.
            z_loc, z_scale = self.encoder(seq_data)
            # Scale log likelihood to account for mini-batching.
            with poutine.scale(scale=local_scale * local_prior_scale):
                # Sample.
                if self.z_prior_distribution == "Normal":
                    pyro.sample("latent", dist.Normal(z_loc, z_scale).to_event(1))
                elif self.z_prior_distribution == "Laplace":
                    pyro.sample("latent", dist.Laplace(z_loc, z_scale).to_event(1))

    def fit_svi(
        self,
        dataset,
        epochs=2,
        anneal_length=1.0,
        batch_size=None,
        scheduler=None,
        jit=False,
    ):
        """
        Infer approximate posterior with stochastic variational inference.

        This runs :class:`~pyro.infer.svi.SVI`. It is an approximate inference
        method useful for quickly iterating on probabilistic models.

        :param ~torch.utils.data.Dataset dataset: The training dataset.
        :param int epochs: Number of epochs of training.
        :param float anneal_length: Number of epochs over which to linearly
            anneal the prior KL divergence weight from 0 to 1, for improved
            training.
        :param int batch_size: Minibatch size (number of sequences).
        :param pyro.optim.MultiStepLR scheduler: Optimization scheduler.
            (Default: Adam optimizer, 0.01 constant learning rate.)
        :param bool jit: Whether to use a jit compiled ELBO.
        """

        # Setup.
        if batch_size is not None:
            self.batch_size = batch_size
        if scheduler is None:
            scheduler = MultiStepLR(
                {
                    "optimizer": Adam,
                    "optim_args": {"lr": 0.01},
                    "milestones": [],
                    "gamma": 0.5,
                }
            )
        if self.is_cuda:
            device = torch.device("cuda")
        else:
            device = torch.device("cpu")
        dataload = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=self.pin_memory,
            generator=torch.Generator(device=device),
        )
        # Initialize guide.
        for seq_data, L_data in dataload:
            if self.is_cuda:
                seq_data = seq_data.cuda()
            self.guide(seq_data, torch.tensor(1.0), torch.tensor(1.0))
            break
        # Setup stochastic variational inference.
        if jit:
            elbo = JitTrace_ELBO(ignore_jit_warnings=True)
        else:
            elbo = Trace_ELBO()
        svi = SVI(self.model, self.guide, scheduler, loss=elbo)

        # Run inference.
        losses = []
        step_i = 1
        t0 = datetime.datetime.now()
        for epoch in range(epochs):
            for seq_data, L_data in dataload:
                if self.is_cuda:
                    seq_data = seq_data.cuda()
                loss = svi.step(
                    seq_data,
                    torch.tensor(len(dataset) / seq_data.shape[0]),
                    self._beta_anneal(step_i, batch_size, len(dataset), anneal_length),
                )
                losses.append(loss)
                scheduler.step()
                step_i += 1
            print(epoch, loss, " ", datetime.datetime.now() - t0)

        return losses

    def _beta_anneal(self, step, batch_size, data_size, anneal_length):
        """Annealing schedule for prior KL term (beta annealing)."""
        if np.allclose(anneal_length, 0.0):
            return torch.tensor(1.0)
        anneal_frac = step * batch_size / (anneal_length * data_size)
        return torch.tensor(min([anneal_frac, 1.0]))

    def evaluate(self, dataset_train, dataset_test=None, jit=False):
        """
        Evaluate performance (log probability and per residue perplexity) on
        train and test datasets.

        :param ~torch.utils.data.Dataset dataset: The training dataset.
        :param ~torch.utils.data.Dataset dataset: The testing dataset
            (optional).
        :param bool jit: Whether to use a jit compiled ELBO.
        """
        dataload_train = DataLoader(dataset_train, batch_size=1, shuffle=False)
        if dataset_test is not None:
            dataload_test = DataLoader(dataset_test, batch_size=1, shuffle=False)
        # Initialize guide.
        for seq_data, L_data in dataload_train:
            if self.is_cuda:
                seq_data = seq_data.cuda()
            self.guide(seq_data, torch.tensor(1.0), torch.tensor(1.0))
            break
        if jit:
            elbo = JitTrace_ELBO(ignore_jit_warnings=True)
        else:
            elbo = Trace_ELBO()
        scheduler = MultiStepLR({"optimizer": Adam, "optim_args": {"lr": 0.01}})
        # Setup stochastic variational inference.
        svi = SVI(self.model, self.guide, scheduler, loss=elbo)

        # Compute elbo and perplexity.
        train_lp, train_perplex = self._evaluate_local_elbo(
            svi, dataload_train, len(dataset_train)
        )
        if dataset_test is not None:
            test_lp, test_perplex = self._evaluate_local_elbo(
                svi, dataload_test, len(dataset_test)
            )
            return train_lp, test_lp, train_perplex, test_perplex
        else:
            return train_lp, None, train_perplex, None

    def _local_variables(self, name, site):
        """Return per datapoint random variables in model."""
        return name in ["latent", "obs_L", "obs_seq"]

    def _evaluate_local_elbo(self, svi, dataload, data_size):
        """Evaluate elbo and average per residue perplexity."""
        lp, perplex = 0.0, 0.0
        with torch.no_grad():
            for seq_data, L_data in dataload:
                if self.is_cuda:
                    seq_data, L_data = seq_data.cuda(), L_data.cuda()
                conditioned_model = poutine.condition(
                    self.model, data={"obs_seq": seq_data}
                )
                args = (seq_data, torch.tensor(1.0), torch.tensor(1.0))
                guide_tr = poutine.trace(self.guide).get_trace(*args)
                model_tr = poutine.trace(
                    poutine.replay(conditioned_model, trace=guide_tr)
                ).get_trace(*args)
                local_elbo = (
                    (
                        model_tr.log_prob_sum(self._local_variables)
                        - guide_tr.log_prob_sum(self._local_variables)
                    )
                    .cpu()
                    .numpy()
                )
                lp += local_elbo
                perplex += -local_elbo / L_data[0].cpu().numpy()
        perplex = np.exp(perplex / data_size)
        return lp, perplex

    def embed(self, dataset, batch_size=None):
        """
        Get the latent space embedding (mean posterior value of z).

        :param ~torch.utils.data.Dataset dataset: The dataset to embed.
        :param int batch_size: Minibatch size (number of sequences). (Defaults
            to batch_size of the model object.)
        """
        if batch_size is None:
            batch_size = self.batch_size
        dataload = DataLoader(dataset, batch_size=batch_size, shuffle=False)
        with torch.no_grad():
            z_locs, z_scales = [], []
            for seq_data, L_data in dataload:
                if self.is_cuda:
                    seq_data = seq_data.cuda()
                z_loc, z_scale = self.encoder(seq_data)
                z_locs.append(z_loc.cpu())
                z_scales.append(z_scale.cpu())

        return torch.cat(z_locs), torch.cat(z_scales)

    def _reconstruct_regressor_seq(self, data, ind, param):
        "Reconstruct the latent regressor sequence given data."
        with torch.no_grad():
            # Encode seq.
            z_loc = self.encoder(data[ind][0])[0]
            # Reconstruct
            decoded = self.decoder(
                z_loc, param("W_q_mn"), param("B_q_mn"), param("inverse_temp_q_mn")
            )
            return torch.exp(decoded["precursor_seq_logits"])
